import os
from typing import Callable, Optional

import numpy as np
from importRosbag.importRosbag import importRosbag
from tonic.dataset import Dataset
from tonic.download_utils import check_integrity, download_url
from tonic.io import make_structured_array


class VPR(Dataset):
    """`Visual Place Recognition <https://zenodo.org/record/4302805>`_

    Event-Based Visual Place Recognition With Ensembles of Temporal Windows.
    Events have (txyp) ordering.

    .. note::  To be able to read this dataset and its GPS files, you will need the `pynmea2` package installed.

    ::

        @article{fischer2020event,
          title={Event-based visual place recognition with ensembles of temporal windows},
          author={Fischer, Tobias and Milford, Michael},
          journal={IEEE Robotics and Automation Letters},
          volume={5},
          number={4},
          pages={6924--6931},
          year={2020},
          publisher={IEEE}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    base_url = "https://zenodo.org/record/4302805/files/"
    recordings = [  # recording names and their md5 hash
        [
            ["dvs_vpr_2020-04-21-17-03-03.bag", "04473f623aec6bda3d7eadfecfc1b2ce"],
            ["20200421_170039-sunset1_concat.nmea", "11f0107a4df845fd315e9134fbee5c1e"],
        ],
        [
            ["dvs_vpr_2020-04-22-17-24-21.bag", "ca6db080a4054196fe65825bce3db351"],
            ["20200422_172431-sunset2_concat.nmea", "ff879bf22a9552a6d8500a98cff6c7f9"],
        ],
        [
            ["dvs_vpr_2020-04-24-15-12-03.bag", "909569732e323ff04c94379a787f2a69"],
            ["20200424_151015-daytime_concat.nmea", "867fdf43ef393ac7e8de251c1a5cd585"],
        ],
        [
            ["dvs_vpr_2020-04-27-18-13-29.bag", "e80b6c0434690908d855445792d4de3b"],
            ["20200427_181204-night_concat.nmea", "441e6673e0dfc8746f76cd646c4aba8d"],
        ],
        [
            ["dvs_vpr_2020-04-28-09-14-11.bag", "7854ede61c0947adb0f072a041dc3bad"],
            ["20200428_091154-morning_concat.nmea", "b86af464ceac478711e52ef4271c198c"],
        ],
        [
            ["dvs_vpr_2020-04-29-06-20-23.bag", "d7ccfeb6539f1e7b077ab4fe6f45193c"],
            ["20200429_061912-sunrise_concat.nmea", "ec04cf35c10eb5b519b11297adef024b"],
        ],
    ]

    sensor_size = (346, 260, 2)
    dtype = np.dtype([("t", int), ("x", int), ("y", int), ("p", int)])
    ordering = dtype.names

    def __init__(
        self,
        save_to,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )

        if not self._check_exists():
            self.download()

    def __getitem__(self, index):
        """
        Returns:
            a tuple of (data, target) where data is another tuple of (events, imu, images) and target is gps positional data.
        """
        (bag_filename, _), (gps_filename, _) = self.recordings[index]
        file_path = os.path.join(self.location_on_system, bag_filename)
        topics = importRosbag(filePathOrName=file_path, log="ERROR")
        events = topics["/dvs/events"]
        events["ts"] -= events["ts"][0]
        events["ts"] *= 1e6
        events = make_structured_array(
            events["ts"], events["x"], events["y"], events["pol"], dtype=self.dtype
        )
        imu = topics["/dvs/imu"]
        imu["ts"] = ((imu["ts"] - imu["ts"][0]) * 1e6).astype(int)
        images = topics["/dvs/image_raw"]
        incorrect_shape_indices = [
            i
            for i, image in enumerate(images["frames"])
            if not (image.shape[0] == 260 and image.shape[1] == 346)
        ]
        for (
            index
        ) in (
            incorrect_shape_indices
        ):  # fix frame shapes that don't match for some reason
            shape_diff_x = self.sensor_size[0] - images["frames"][index].shape[1]
            shape_diff_y = self.sensor_size[1] - images["frames"][index].shape[0]
            images["frames"][index] = np.pad(
                images["frames"][index], [(0, shape_diff_y), (0, shape_diff_x), (0, 0)]
            )
        images["frames"] = np.stack(images["frames"])  # errors for some recordings
        images["ts"] = ((images["ts"] - images["ts"][0]) * 1e6).astype(int)
        data = events, imu, images

        targets = self.read_gps_file(
            os.path.join(self.location_on_system, gps_filename)
        )

        if self.transform is not None:
            data = self.transform(data)
        if self.target_transform is not None:
            targets = self.target_transform(targets)
        if self.transforms is not None:
            data, targets = self.transforms(data, targets)
        return data, targets

    def __len__(self):
        return len(self.recordings)

    def download(self):
        for recording in self.recordings:
            for filename, md5_hash in recording:
                download_url(
                    self.base_url + filename,
                    self.location_on_system,
                    filename=filename,
                    md5=md5_hash,
                )

    def _check_exists(self):
        # check if all filenames are correct
        files_present = list(
            [
                check_integrity(os.path.join(self.location_on_system, filename))
                for recording in self.recordings
                for filename, md5 in recording
            ]
        )
        return all(files_present)

    # code taken from https://github.com/Tobias-Fischer/ensemble-event-vpr/blob/master/read_gps.py
    def read_gps_file(self, nmea_file_path):
        import pynmea2

        nmea_file = open(nmea_file_path, encoding="utf-8")
        latitudes, longitudes, timestamps = [], [], []
        first_timestamp = None
        previous_lat, previous_lon = 0, 0

        for line in nmea_file.readlines():
            try:
                msg = pynmea2.parse(line)
                if first_timestamp is None:
                    first_timestamp = msg.timestamp
                if msg.sentence_type not in ["GSV", "VTG", "GSA"]:
                    dist_to_prev = np.linalg.norm(
                        np.array([msg.latitude, msg.longitude])
                        - np.array([previous_lat, previous_lon])
                    )
                    if (
                        msg.latitude != 0
                        and msg.longitude != 0
                        and msg.latitude != previous_lat
                        and msg.longitude != previous_lon
                        and dist_to_prev > 0.0001
                    ):
                        timestamp_diff = (
                            (msg.timestamp.hour - first_timestamp.hour) * 3600
                            + (msg.timestamp.minute - first_timestamp.minute) * 60
                            + (msg.timestamp.second - first_timestamp.second)
                        )
                        latitudes.append(msg.latitude)
                        longitudes.append(msg.longitude)
                        timestamps.append(timestamp_diff)
                        previous_lat, previous_lon = msg.latitude, msg.longitude

            except pynmea2.ParseError as e:
                continue

        return np.array(np.vstack((latitudes, longitudes, timestamps))).T



import numpy as np
import tonic.transforms as transforms


def plot_event_grid(events, axis_array=(1, 3), plot_frame_number=False):
    """Plot events accumulated as frames equal to the product
    of axes for visual inspection.

    Parameters:
        events: Structured numpy array of shape [num_events, num_event_channels].
        axis_array: dimensions of plotting grid. The larger the grid,
                    the more fine-grained the events will be sliced in time.
        plot_frame_number: optional index of frame when plotting

    Example:
        >>> import tonic
        >>> dataset = tonic.datasets.NMNIST(save_to='./data')
        >>> events, target = dataset[100]
        >>> tonic.utils.plot_event_grid(events)

    Returns:
        None
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError(
            "Please install the matplotlib package to plot events. This is an optional"
            " dependency."
        )

    if "y" in events.dtype.names:
        sensor_size_x = int(events["x"].max() + 1)
        sensor_size_y = int(events["y"].max() + 1)
        sensor_size_p = len(np.unique(events["p"]))
        sensor_size = (sensor_size_x, sensor_size_y, sensor_size_p)

        transform = transforms.ToFrame(
            sensor_size=sensor_size, n_time_bins=np.product(axis_array)
        )

        frames = transform(events)
        fig, axes_array = plt.subplots(*axis_array)

        if 1 in axis_array:
            axes_array = axes_array.reshape(1, -1)

        for i in range(axis_array[0]):
            for j in range(axis_array[1]):
                frame = frames[i * axis_array[1] + j]
                axes_array[i, j].imshow(frame[1] - frame[0])
                axes_array[i, j].axis("off")
                if plot_frame_number:
                    axes_array[i, j].title.set_text(str(i * axis_array[1] + j))
        plt.tight_layout()
        plt.show()

    else:
        sensor_size_x = int(events["x"].max() + 1)
        frame_transform = transforms.ToFrame(
            sensor_size=(sensor_size_x, 1, 1), n_time_bins=sensor_size_x * 2
        )

        frames = frame_transform(events)
        plt.imshow(frames.squeeze().T)
        plt.xlabel("Time")
        plt.ylabel("Channels")


def plot_animation(frames: np.ndarray):
    """Helper function that animates a tensor of frames of shape (TCHW).

    Parameters:
        frames: numpy array or tensor of shape (TCHW)

    Example:
        >>> import tonic
        >>> nmnist = tonic.datasets.NMNIST(save_to='./data', train=False)
        >>> events, label = nmnist[0]
        >>>
        >>> transform = tonic.transforms.ToFrame(
        >>>     sensor_size=nmnist.sensor_size,
        >>>     time_window=10000,
        >>> )
        >>>
        >>> frames = transform(events)
        >>> animation = tonic.utils.plot_animation(frames)

    Returns:
        The animation object. Store this in a variable to keep it from being garbage collected until displayed.
    """
    try:
        from matplotlib import animation
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError(
            "Please install the matplotlib package to plot events. This is an optional"
            " dependency."
        )
    fig = plt.figure(figsize=(2, 2))
    if frames[0].shape[0] == 2:
        first_frame = frames[0][1] - frames[0][0]
    else:
        first_frame = frames[0][0]
    ax = plt.imshow(first_frame)
    plt.axis("off")

    def animate(frame):
        if frame.shape[0] == 2:
            frame = frame[1] - frame[0]
        else:
            frame = frame[0]
        ax.set_data(frame)
        return ax

    anim = animation.FuncAnimation(fig, animate, frames=frames, interval=100)
    plt.show()
    return anim



from .crop import crop_numpy
from .decimate import decimate_numpy
from .denoise import denoise_numpy
from .drop_event import drop_event_numpy, drop_by_area_numpy, drop_by_time_numpy
from .drop_pixel import (
    drop_pixel_numpy,
    drop_pixel_raster,
    identify_hot_pixel,
    identify_hot_pixel_raster,
)
from .refractory_period import refractory_period_numpy
from .spatial_jitter import spatial_jitter_numpy
from .time_jitter import time_jitter_numpy
from .time_skew import time_skew_numpy
from .to_averaged_timesurface import to_averaged_timesurface
from .to_frame import to_frame_numpy
from .to_timesurface import to_timesurface_numpy
from .to_voxel_grid import to_voxel_grid_numpy
from .to_bina_rep import to_bina_rep_numpy
from .uniform_noise import uniform_noise_numpy

__all__ = [
    "crop_numpy",
    "decimate_numpy",
    "denoise_numpy",
    "drop_event_numpy",
    "drop_by_area_numpy",
    "drop_by_time_numpy",
    "drop_pixel_numpy",
    "refractory_period_numpy",
    "spatial_jitter_numpy",
    "spatial_resize_numpy",
    "time_jitter_numpy",
    "time_skew_numpy",
    "to_averaged_timesurface",
    "to_frame_numpy",
    "to_timesurface_numpy",
    "to_voxel_grid_numpy",
    "to_bina_rep_numpy",
    "uniform_noise_numpy",
]



import os
from typing import Callable, Optional

import numpy as np
from tonic.dataset import Dataset
from tonic.download_utils import extract_archive


class DVSLip(Dataset):
    """`DVS-Lip <https://sites.google.com/view/event-based-lipreading>`_
    ::

        @inproceedings{tan2022multi,
            title={Multi-Grained Spatio-Temporal Features Perceived Network for Event-Based Lip-Reading},
            author={Tan, Ganchao and Wang, Yang and Han, Han and Cao, Yang and Wu, Feng and Zha, Zheng-Jun},
            booktitle={Proceedings of the IEEE/CVF Conference on Computer Vision and Pattern Recognition},
            pages={20094--20103},
            year={2022}
        }

        Implementation inspired from original script: https://github.com/tgc1997/event-based-lip-reading/blob/main/utils/dataset.py

    Parameters:
        save_to (string): Location to save files to on disk.
        train (bool): If True, uses training subset, otherwise testing subset.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    base_url = "https://drive.google.com/file/d/1dBEgtmctTTWJlWnuWxFtk8gfOdVVpkQ0/view"
    filename = "DVS-Lip.zip"
    base_folder = "DVS-Lip"
    file_md5 = "2dcb959255122d4cdeb6094ca282494b"

    sensor_size = (128, 128, 2)
    dtype = np.dtype([("x", np.int16), ("y", np.int16), ("p", bool), ("t", np.int64)])
    ordering = dtype.names

    classes = [
        "accused",
        "action",
        "allow",
        "allowed",
        "america",
        "american",
        "another",
        "around",
        "attacks",
        "banks",
        "become",
        "being",
        "benefit",
        "benefits",
        "between",
        "billion",
        "called",
        "capital",
        "challenge",
        "change",
        "chief",
        "couple",
        "court",
        "death",
        "described",
        "difference",
        "different",
        "during",
        "economic",
        "education",
        "election",
        "england",
        "evening",
        "everything",
        "exactly",
        "general",
        "germany",
        "giving",
        "ground",
        "happen",
        "happened",
        "having",
        "heavy",
        "house",
        "hundreds",
        "immigration",
        "judge",
        "labour",
        "leaders",
        "legal",
        "little",
        "london",
        "majority",
        "meeting",
        "military",
        "million",
        "minutes",
        "missing",
        "needs",
        "number",
        "numbers",
        "paying",
        "perhaps",
        "point",
        "potential",
        "press",
        "price",
        "question",
        "really",
        "right",
        "russia",
        "russian",
        "saying",
        "security",
        "several",
        "should",
        "significant",
        "spend",
        "spent",
        "started",
        "still",
        "support",
        "syria",
        "syrian",
        "taken",
        "taking",
        "terms",
        "these",
        "thing",
        "think",
        "times",
        "tomorrow",
        "under",
        "warning",
        "water",
        "welcome",
        "words",
        "worst",
        "years",
        "young",
    ]  # 100 labels

    ambiguous_classes = [
        "action",
        "allow",
        "allowed",
        "america",
        "american",
        "around",
        "being",
        "benefit",
        "benefits",
        "billion",
        "called",
        "challenge",
        "change",
        "court",
        "difference",
        "different",
        "election",
        "evening",
        "giving",
        "ground",
        "happen",
        "happened",
        "having",
        "heavy",
        "legal",
        "little",
        "meeting",
        "million",
        "missing",
        "needs",
        "number",
        "numbers",
        "paying",
        "press",
        "price",
        "russia",
        "russian",
        "spend",
        "spent",
        "syria",
        "syrian",
        "taken",
        "taking",
        "terms",
        "these",
        "thing",
        "think",
        "times",
        "words",
        "worst",
    ]  # the 25 word pairs that are ambiguous (see paper)

    def __init__(
        self,
        save_to: str,
        train: bool = True,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )
        self.train = train
        self.url = self.base_url
        self.folder_name = os.path.join(
            self.base_folder, "train" if self.train else "test"
        )

        if not self._check_exists():
            if self._is_file_present():  # check if zip file is manually downloaded
                archive = os.path.join(self.location_on_system, self.filename)
                print(f"Extracting {archive} to {self.location_on_system}...")
                extract_archive(
                    archive,
                    remove_finished=True,
                )
            else:
                print(
                    f"""
                    WARNING: this dataset is available from Google Drive and must be downloaded manually.
                    Please download the zip file ( {self.url} ) and place it in {self.location_on_system}."""
                )
                exit()

        file_path = os.path.join(self.location_on_system, self.folder_name)

        for act_dir in os.listdir(file_path):
            label = self.classes.index(act_dir)

            for file in os.listdir(os.path.join(file_path, act_dir)):
                if file.endswith("npy"):
                    self.targets.append(label)
                    self.data.append(os.path.join(file_path, act_dir, file))

    def __getitem__(self, index):
        """
        Returns:
            a tuple of (events, target) where target is the index of the target class.
        """
        orig_events = np.load(self.data[index])

        events = np.empty(orig_events.shape, dtype=self.dtype)
        events["x"] = orig_events["x"]
        events["y"] = orig_events["y"]
        events["t"] = orig_events["t"]
        events["p"] = orig_events["p"]

        target = self.targets[index]
        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.transforms is not None:
            events, target = self.transforms(events, target)
        return events, target

    def __len__(self):
        return len(self.data)

    def _check_exists(self):
        return os.path.isdir(
            os.path.join(
                self.location_on_system, self.folder_name
            )  # check if directory exists
        ) and self._folder_contains_at_least_n_files_of_type(100, ".npy")



import tonic

project = "Tonic"
copyright = "2019-present, the neuromorphs of Telluride"
author = "Gregor Lenz"

master_doc = "index"

extensions = [
    "myst_nb",
    "pbr.sphinxext",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_gallery.gen_gallery",
]

sphinx_gallery_conf = {
    "examples_dirs": "gallery/",  # path to your example scripts
    "gallery_dirs": "auto_examples",  # path to where to save gallery generated output
    # "backreferences_dir": "gen_modules/backreferences",
    "matplotlib_animations": True,
    "doc_module": ("tonic",),
    # "binder": True,
    "download_all_examples": False,
    "ignore_pattern": r"utils\.py",
}

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True

# MyST settings
# nb_execution_mode = "off"
nb_execution_timeout = 300
nb_execution_show_tb = True
nb_execution_excludepatterns = ["large_datasets.ipynb"]
suppress_warnings = ["myst.header"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
html_theme = "sphinx_book_theme"
html_title = tonic.__version__
html_logo = "_static/tonic-logo-black.png"
html_favicon = "_static/tonic_favicon.png"
html_show_sourcelink = True
html_sourcelink_suffix = ""

html_theme_options = {
    "repository_url": "https://github.com/neuromorphs/tonic",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
    "repository_branch": "develop",
    "path_to_docs": "docs",
    "use_fullscreen_button": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



import numpy as np
from typing import List


def spatial_jitter_numpy(
    events: np.ndarray,
    sensor_size: List[int],
    var_x: float = 1,
    var_y: float = 1,
    sigma_xy: float = 0,
    clip_outliers: bool = False,
):
    """Changes x/y coordinate for each event by adding samples from a multivariate
    Gaussian distribution. It with the following properties:

        .. math::
            mean = [x,y]

            \Sigma = [[var_x, sigma_{xy}],[sigma_{xy}, var_y]]

    Jittered events that lie outside the focal plane will be dropped if clip_outliers is True.

    Parameters:
        events: ndarray of shape [num_events, num_event_channels]
        variance_x: squared sigma value for the distribution in the x direction
        variance_y: squared sigma value for the distribution in the y direction
        sigma_xy: changes skewness of distribution, only change if you want shifts along diagonal axis.
        clip_outliers: when True, events that have been jittered outside the sensor size will be dropped.

    Returns:
        array of spatially jittered events.
    """

    assert "x" and "y" in events.dtype.names

    shifts = np.random.multivariate_normal(
        [0, 0], [[var_x, sigma_xy], [sigma_xy, var_y]], len(events)
    )

    events["x"] = events["x"] + shifts[:, 0]
    events["y"] = events["y"] + shifts[:, 1]

    if clip_outliers:
        events = np.delete(
            events,
            np.where(
                (events["x"] < 0)
                | (events["x"] >= sensor_size[0])
                | (events["y"] < 0)
                | (events["y"] >= sensor_size[1])
            ),
            axis=0,
        )

    return events



"""
=============
ToVoxelGrid
=============

:class:`~tonic.transforms.ToVoxelGrid`
"""

import tonic

nmnist = tonic.datasets.NMNIST("../../tutorials/data", train=False)
events, label = nmnist[0]

transform = tonic.transforms.ToVoxelGrid(
    sensor_size=nmnist.sensor_size,
    n_time_bins=20,
)

frames = transform(events)

ani = tonic.utils.plot_animation(frames)



import numpy as np


# from https://gitlab.com/synsense/aermanager/-/blob/master/aermanager/preprocess.py#L188
def identify_hot_pixel(events: np.ndarray, hot_pixel_frequency: float):
    """Identifies pixels that fire above above a certain frequency, averaged across
    whole event recording. Such _hot_ pixels are sometimes caused by faulty hardware.

    Parameters:
        events: ndarray of shape [num_events, num_event_channels]
        hot_pixel_frequency: number of spikes per pixel allowed for the recording, any pixel
                             firing above that number will be deactivated.

    Returns:
        list of (x/y) coordinates for excessively firing pixels.
    """

    assert "x" and "y" and "t" in events.dtype.names

    total_time = events["t"][-1] - events["t"][0]

    hist = np.histogram2d(
        events["x"],
        events["y"],
        bins=(np.arange(events["y"].max() + 1), np.arange(events["x"].max() + 1)),
    )[0]
    max_occur = hot_pixel_frequency * total_time * 1e-6
    hot_pixels = np.asarray((hist > max_occur).nonzero()).T

    return hot_pixels


def identify_hot_pixel_raster(events: np.ndarray, hot_pixel_frequency: float):
    """Identifies pixels that fire above a certain predefined spike amount, supports both

    Parameters:
        events: ndarray of shape [P, H, W] or [T, P, H, W]
        hot_pixel_frequency: number of spikes per pixel allowed for the recording, any pixel
                             firing above that number will be deactivated.

    Returns:
        list of (x/y) coordinates for excessively firing pixels.
    """
    if len(events.shape) == 3:
        # if dim =3, the input is frame
        merged_polarity = events.copy().sum(0)
    elif len(events.shape) == 4:
        # if dimension is 4, input is raster
        merged_polarity = events.copy().sum(0).sum(0)

    ind = np.argwhere(merged_polarity > hot_pixel_frequency)
    a = 1
    return tuple(zip(ind[:, 0], ind[:, 1]))


def drop_pixel_numpy(events: np.ndarray, coordinates):
    """Drops events for pixel locations that fire

    Parameters:
        events: ndarray of shape [num_events, num_event_channels]
        ordering: ordering of the event tuple inside of events. This function requires 'x' and
                  'y' to be in the ordering
        coordinates: list of (x,y) coordinates for which all events will be deleted.

    Returns:
        subset of original events.
    """

    assert "x" and "y" in events.dtype.names

    dropped_pixel_mask = np.full((events.shape[0]), False, dtype=bool)
    for x, y in coordinates:
        current_mask = np.logical_and(events["x"] == x, events["y"] == y)
        dropped_pixel_mask = np.logical_or(current_mask, dropped_pixel_mask)

    return events[np.invert(dropped_pixel_mask)]


def drop_pixel_raster(raster: np.ndarray, coordinates):
    """Drops events for pixel locations

    Parameters:
        events: ndarray of shape [p, h, w] or [t, p, h, w]
        ordering: ordering of the event tuple inside of events. This function requires 'x' and
                  'y' to be in the ordering
        coordinates: list of (x,y) coordinates for which all events will be deleted.

    Returns:
        The filtered raster or frame
    """
    assert len(raster.shape) == 4 or len(raster.shape) == 3

    for x, y in coordinates:
        raster[..., x, y] = 0

    return raster



"""
==================
SpatialJitter
==================
The :class:`~tonic.transforms.SpatialJitter` jitters
x and y coordinates of events.
"""

import tonic

nmnist = tonic.datasets.NMNIST("../../tutorials/data", train=False)
events, label = nmnist[0]

transform = tonic.transforms.Compose(
    [
        tonic.transforms.SpatialJitter(
            sensor_size=nmnist.sensor_size, var_x=1, var_y=1, clip_outliers=True
        ),
        tonic.transforms.ToFrame(
            sensor_size=nmnist.sensor_size,
            time_window=10000,
        ),
    ]
)

frames = transform(events)

ani = tonic.utils.plot_animation(frames)



import numpy as np
from typing import Tuple


def uniform_noise_numpy(events: np.ndarray, sensor_size: Tuple[int, int, int], n: int):
    """
    Adds a fixed number of noise events that are uniformly distributed across sensor size dimensions.

    Parameters:
        events: ndarray of shape (n_events, n_event_channels)
        sensor_size: 3-tuple of integers for x, y, p
        n: the number of noise events added.
    """
    noise_events = np.zeros(n, dtype=events.dtype)
    for channel in events.dtype.names:
        if channel == "x":
            low, high = 0, sensor_size[0]
        if channel == "y":
            low, high = 0, sensor_size[1]
        if channel == "p":
            low, high = 0, sensor_size[2]
        if channel == "t":
            low, high = events["t"].min(), events["t"].max()
        noise_events[channel] = np.random.uniform(low=low, high=high, size=n)
    noisy_events = np.concatenate((events, noise_events))
    return noisy_events[np.argsort(noisy_events["t"])]



import os
from typing import Callable, Optional

import h5py
import numpy as np
from tonic.dataset import Dataset
from tonic.io import make_structured_array


class HSD(Dataset):
    """Heidelberg Spiking Dataset <https://arxiv.org/abs/1910.07407> contains the Spiking Heidelberg Dataset (SHD)
    and the Spiking Speech Commands dataset (SSC)."""

    base_url = "https://zenkelab.org/datasets/"
    sensor_size = (700, 1, 1)
    dtype = np.dtype([("t", int), ("x", int), ("p", int)])
    ordering = dtype.names

    def __getitem__(self, index):
        file = h5py.File(os.path.join(self.location_on_system, self.data_filename), "r")
        # adding artificial polarity of 1 and convert to microseconds
        events = make_structured_array(
            file["spikes/times"][index] * 1e6,
            file["spikes/units"][index],
            1,
            dtype=self.dtype,
        )
        target = file["labels"][index].astype(int)
        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        file = h5py.File(os.path.join(self.location_on_system, self.data_filename), "r")
        return len(file["labels"])

    def _check_exists(self):
        return (
            self._is_file_present()
            and self._folder_contains_at_least_n_files_of_type(1, ".h5")
        )


class SHD(HSD):
    """`Spiking Heidelberg Digits <https://arxiv.org/abs/1910.07407>`_
    ::

        @article{cramer2020heidelberg,
          title={The heidelberg spiking data sets for the systematic evaluation of spiking neural networks},
          author={Cramer, Benjamin and Stradmann, Yannik and Schemmel, Johannes and Zenke, Friedemann},
          journal={IEEE Transactions on Neural Networks and Learning Systems},
          year={2020},
          publisher={IEEE}
        }

    Parameters:
        save_to (string): Location to save files to on disk. Will put files in an 'hsd' subfolder.
        train (bool): If True, uses training subset, otherwise testing subset.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.

    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, targets).
    """

    test_zip = "shd_test.h5.zip"
    train_zip = "shd_train.h5.zip"
    test_md5 = "1503a5064faa34311c398fb0a1ed0a6f"
    train_md5 = "f3252aeb598ac776c1b526422d90eecb"
    folder_name = ""

    def __init__(
        self,
        save_to: str,
        train: bool = True,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
        )

        if train:
            self.url = self.base_url + self.train_zip
            self.filename = self.train_zip
            self.file_md5 = self.train_md5
        else:
            self.url = self.base_url + self.test_zip
            self.filename = self.test_zip
            self.file_md5 = self.test_md5
        self.data_filename = self.filename[:-4]

        if not self._check_exists():
            self.download()

        file = h5py.File(os.path.join(self.location_on_system, self.data_filename), "r")
        self.classes = file["extra/keys"][()]


class SSC(HSD):
    """`Spiking Speech Commands <https://arxiv.org/abs/1910.07407>`_
    ::

        @article{cramer2020heidelberg,
          title={The heidelberg spiking data sets for the systematic evaluation of spiking neural networks},
          author={Cramer, Benjamin and Stradmann, Yannik and Schemmel, Johannes and Zenke, Friedemann},
          journal={IEEE Transactions on Neural Networks and Learning Systems},
          year={2020},
          publisher={IEEE}
        }

    Parameters:
        save_to (string): Location to save files to on disk. Will put files in an 'hsd' subfolder.
        split (string): One of 'train', 'test' or 'valid'.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.

    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, targets).
    """

    test_zip = "ssc_test.h5.zip"
    train_zip = "ssc_train.h5.zip"
    valid_zip = "ssc_valid.h5.zip"
    test_md5 = "a35ff1e9cffdd02a20eb850c17c37748"
    train_md5 = "d102be95e7144fcc0553d1f45ba94170"
    valid_md5 = "b4eee3516a4a90dd0c71a6ac23a8ae43"
    folder_name = ""

    def __init__(
        self,
        save_to: str,
        split: str = "train",
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ):
        super().__init__(
            save_to, transform=transform, target_transform=target_transform
        )

        if split == "train":
            self.url = self.base_url + self.train_zip
            self.filename = self.train_zip
            self.file_md5 = self.train_md5
        elif split == "test":
            self.url = self.base_url + self.test_zip
            self.filename = self.test_zip
            self.file_md5 = self.test_md5
        elif split == "valid":
            self.url = self.base_url + self.valid_zip
            self.filename = self.valid_zip
            self.file_md5 = self.valid_md5
        self.data_filename = self.filename[:-4]

        if not self._check_exists():
            self.download()

        file = h5py.File(os.path.join(self.location_on_system, self.data_filename), "r")
        self.classes = file["extra/keys"][()]



import os
from typing import Callable, Optional

import numpy as np
from tonic.dataset import Dataset
from tonic.io import read_mnist_file


class NMNIST(Dataset):
    """`N-MNIST <https://www.garrickorchard.com/datasets/n-mnist>`_

    Events have (xytp) ordering.
    ::

        @article{orchard2015converting,
          title={Converting static image datasets to spiking neuromorphic datasets using saccades},
          author={Orchard, Garrick and Jayawant, Ajinkya and Cohen, Gregory K and Thakor, Nitish},
          journal={Frontiers in neuroscience},
          volume={9},
          pages={437},
          year={2015},
          publisher={Frontiers}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        train (bool): If True, uses training subset, otherwise testing subset.
        first_saccade_only (bool): If True, only work with events of the first of three saccades.
                                   Results in about a third of the events overall.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    base_url = "https://data.mendeley.com/public-files/datasets/468j46mzdv/files/"
    train_url = base_url + "39c25547-014b-4137-a934-9d29fa53c7a0/file_downloaded"
    train_filename = "train.zip"
    train_md5 = "20959b8e626244a1b502305a9e6e2031"
    train_folder = "Train"
    test_url = base_url + "05a4d654-7e03-4c15-bdfa-9bb2bcbea494/file_downloaded"
    test_filename = "test.zip"
    test_md5 = "69ca8762b2fe404d9b9bad1103e97832"
    test_folder = "Test"

    classes = [
        "0 - zero",
        "1 - one",
        "2 - two",
        "3 - three",
        "4 - four",
        "5 - five",
        "6 - six",
        "7 - seven",
        "8 - eight",
        "9 - nine",
    ]

    sensor_size = (34, 34, 2)
    dtype = np.dtype([("x", int), ("y", int), ("t", int), ("p", int)])
    ordering = dtype.names

    def __init__(
        self,
        save_to: str,
        train: bool = True,
        first_saccade_only: bool = False,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )
        self.train = train
        self.first_saccade_only = first_saccade_only

        if train:
            self.filename = self.train_filename
            self.url = self.train_url
            self.file_md5 = self.train_md5
            self.folder_name = self.train_folder
        else:
            self.filename = self.test_filename
            self.url = self.test_url
            self.file_md5 = self.test_md5
            self.folder_name = self.test_folder

        if not self._check_exists():
            self.download()

        file_path = os.path.join(self.location_on_system, self.folder_name)
        for path, dirs, files in os.walk(file_path):
            files.sort()
            for file in files:
                if file.endswith("bin"):
                    self.data.append(path + "/" + file)
                    label_number = int(path[-1])
                    self.targets.append(label_number)

    def __getitem__(self, index):
        """
        Returns:
            a tuple of (events, target) where target is the index of the target class.
        """
        events = read_mnist_file(self.data[index], dtype=self.dtype)
        if self.first_saccade_only:
            events = events[events["t"] < 1e5]
        target = self.targets[index]
        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.transforms is not None:
            events, target = self.transforms(events, target)
        return events, target

    def __len__(self) -> int:
        return len(self.data)

    def _check_exists(self) -> bool:
        return (
            self._is_file_present()
            and self._folder_contains_at_least_n_files_of_type(10000, ".bin")
        )



import os
from struct import unpack

import numpy as np
from tonic.dataset import Dataset
from tonic.download_utils import download_and_extract_archive
from tonic.io import make_structured_array


class SMNIST(Dataset):
    """Spiking sequential `MNIST <http://yann.lecun.com/exdb/mnist/>`_

    Sequential MNIST (sMNIST) is a standard benchmark task for time series
    classification where each input consists of sequences of 784 pixel
    values created by unrolling the MNIST digits, pixel by pixel. In this
    spiking version, each of the 99 input neurons is associated with a
    particular threshold for the grey value, and this input neuron fires
    whenever the grey value crosses its threshold in the transition from
    the previous to the current pixel.

    Parameters:
        save_to (string): Location to save files to on disk.
        train (bool): If True, uses training subset, otherwise testing subset.
        duplicate (bool): If True, emits two spikes per threshold crossing
        num_neurons (integer): How many neurons to use to encode thresholds(must be odd)
        dt (float): Duration(in microseconds) of each timestep
        download (bool): Choose to download data or verify existing files. If True
                        and a file with the same name and correct hash is already
                        in the directory, download is automatically skipped.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.

     Returns:
         A dataset object that can be indexed or iterated over.
         One sample returns a tuple of (events, targets).
    """

    base_url = "https://storage.googleapis.com/cvdf-datasets/mnist/"
    train_images_file = "train-images-idx3-ubyte"
    train_labels_file = "train-labels-idx1-ubyte"
    test_images_file = "t10k-images-idx3-ubyte"
    test_labels_file = "t10k-labels-idx1-ubyte"
    dtype = np.dtype([("t", int), ("x", int), ("p", int)])
    ordering = dtype.names

    classes = [
        "0 - zero",
        "1 - one",
        "2 - two",
        "3 - three",
        "4 - four",
        "5 - five",
        "6 - six",
        "7 - seven",
        "8 - eight",
        "9 - nine",
    ]

    def __init__(
        self,
        save_to,
        train=True,
        duplicate=True,
        num_neurons=99,
        dt=1000.0,
        download=True,
        transform=None,
        target_transform=None,
    ):
        super(SMNIST, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )
        self.location_on_system = os.path.join(save_to, "smnist")
        self.train = train
        self.duplicate = duplicate
        self.sensor_size = (num_neurons, 1, 1)
        self.dt = dt

        if (num_neurons % 2) == 0:
            raise Exception("Number of neurons must be odd")

        self.images_file = self.train_images_file if train else self.test_images_file
        self.labels_file = self.train_labels_file if train else self.test_labels_file

        if download:
            self.download()

        # Open images file
        with open(os.path.join(self.location_on_system, self.images_file), "rb") as f:
            image_data = f.read()

            # Unpack header from first 16 bytes of buffer and verify
            magic, num_items, num_rows, num_cols = unpack(">IIII", image_data[:16])
            assert magic == 2051
            assert num_rows == 28
            assert num_cols == 28

            # Convert remainder of buffer to numpy bytes
            self.image_data = np.frombuffer(image_data[16:], dtype=np.uint8)

            # Reshape data into individual (flattened) images
            self.image_data = np.reshape(self.image_data, (num_items, 28 * 28))

        # Open labels file
        with open(os.path.join(self.location_on_system, self.labels_file), "rb") as f:
            label_data = f.read()

            # Unpack header from first 8 bytes of buffer and verify
            magic, num_items = unpack(">II", label_data[:8])
            assert magic == 2049

            # Convert remainder of buffer to numpy bytes
            self.label_data = np.frombuffer(label_data[8:], dtype=np.uint8)
            assert self.label_data.shape == (self.image_data.shape[0],)

    def __getitem__(self, index):
        image = self.image_data[index]

        # Determine how many neurons should encode onset and offset
        half_size = self.sensor_size[0] // 2

        # Determine thresholds of each neuron
        thresholds = np.linspace(0.0, 254.0, half_size).astype(np.uint8)

        # Determine for which pixels each neuron is above or below its threshol
        lower = image[:, None] < thresholds[None, :]
        higher = image[:, None] >= thresholds[None, :]

        # Get onsets and offset (transitions between lower and higher) spike times and ids
        on_spike_time, on_spike_idx = np.where(np.logical_and(lower[:-1], higher[1:]))
        off_spike_time, off_spike_idx = np.where(np.logical_and(higher[:-1], lower[1:]))
        off_spike_idx += half_size

        # Get times when image is 255 and create matching neuron if
        touch_spike_time = np.where(image == 255)[0]
        touch_spike_idx = (
            np.ones(touch_spike_time.shape, dtype=np.int64) * self.sensor_size[0]
        )

        # Combine all spike times and ids together
        spike_time = np.concatenate((on_spike_time, off_spike_time, touch_spike_time))
        spike_idx = np.concatenate((on_spike_idx, off_spike_idx, touch_spike_idx))
        spike_idx = self.sensor_size[0] - spike_idx

        # Sort, first by spike time and then by spike index
        spike_order = np.lexsort((spike_idx, spike_time))
        spike_time = spike_time[spike_order]
        spike_idx = spike_idx[spike_order]

        # If we should duplicate each spike
        if self.duplicate:
            # Repeat spike indices
            spike_idx = np.repeat(spike_idx, 2)

            # Double spike times
            double_spike_time = spike_time * 2

            # Interleave
            spike_time = np.empty(2 * double_spike_time.shape[0], dtype=np.int64)
            spike_time[0::2] = double_spike_time
            spike_time[1::2] = double_spike_time + 1

        # stack and add artificial polarity of 1
        events = make_structured_array(
            spike_time * self.dt, spike_idx, 1, dtype=self.dtype
        )
        target = self.label_data[index]

        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        return self.image_data.shape[0]

    def download(self):
        for f in [self.images_file, self.labels_file]:
            download_and_extract_archive(
                self.base_url + f + ".gz", self.location_on_system, filename=f + ".gz"
            )



import pytest
import numpy as np


def test_standardize_data_length():
    from tonic.audio_transforms import FixLength

    sdl = FixLength(100, 1)

    # Data is longer
    data = np.ones((1, 120))
    assert sdl(data).shape == (1, 100)

    # Data is shorter
    data = np.ones((1, 80))
    assert sdl(data).shape == (1, 100)


def test_bin():
    import numpy as np
    from tonic.audio_transforms import Bin

    bin_transform = Bin(orig_freq=16_000, new_freq=100, axis=1)

    data = np.random.random((1, 8 * 16_000))
    data_binned = bin_transform(data)

    assert data_binned.shape == (1, 8 * 100)
    assert data.sum() == pytest.approx(data_binned.sum(), 1e-7)


def test_linear_butter_filter_bank():
    from tonic.audio_transforms import LinearButterFilterBank
    import numpy as np

    fb = LinearButterFilterBank(
        order=2, low_freq=100, sampling_freq=16000, num_filters=16
    )
    data = np.random.random((1, 16_000))

    filter_out = fb(data)
    assert filter_out.shape == (16, 16_000)


def test_mel_butter_filter_bank():
    from tonic.audio_transforms import MelButterFilterBank
    import numpy as np

    fb = MelButterFilterBank(order=2, low_freq=100, sampling_freq=16000, num_filters=16)
    data = np.random.random((1, 16_000))

    filter_out = fb(data)
    assert filter_out.shape == (16, 16_000)


def test_add_noise():
    import numpy as np

    class DummyNoiseDataset:
        def __len__(self):
            return 1000

        def __getitem__(self, item):
            sig_len = np.random.randint(12000, 20000, (1,)).item()
            return np.random.random((1, sig_len)), 0

    data = np.sin(np.arange(0, 16_000 * 0.001, 0.001))[None, ...]
    print(data.shape)

    from tonic.audio_transforms import AddNoise

    noise_dataset = DummyNoiseDataset()

    print(noise_dataset[0][0].shape)

    add_noise = AddNoise(noise_dataset, 10, normed=True)

    signal = add_noise(data)
    assert signal.shape == (1, 16_000)



import numpy as np
import warnings


def time_skew_numpy(events: np.ndarray, coefficient: float, offset: int = 0):
    """Skew all event timestamps according to a linear transform,
    potentially sampled from a distribution of acceptable functions.

    Parameters:
        events: ndarray of shape [num_events, num_event_channels].
        coefficient: a real-valued multiplier applied to the timestamps of the events.
                     E.g. a coefficient of 2.0 will double the effective delay between any
                     pair of events.
        offset: value by which the timestamps will be shifted after multiplication by
                the coefficient. Negative offsets are permissible but may result in
                in an exception if timestamps are shifted below 0.

    Returns:
        the input events with rewritten timestamps.
    """

    assert "t" in events.dtype.names

    if isinstance(coefficient, tuple):
        coefficient = (
            coefficient[1] - coefficient[0]
        ) * np.random.random_sample() + coefficient[0]

    if isinstance(offset, tuple):
        offset = (offset[1] - offset[0]) * np.random.random_sample() + offset[0]

    events["t"] = events["t"] * coefficient + offset

    return events



import numpy as np


def to_timesurface_numpy(
    events, sensor_size, surface_dimensions=None, tau=5e3, decay="lin"
):
    """Representation that creates timesurfaces for each event in the recording. Modeled after the paper
    Lagorce et al. 2016, Hots: a hierarchy of event-based time-surfaces for pattern recognition
    https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=7508476

    Parameters:
        surface_dimensions (int, int): width does not have to be equal to height, however both numbers have to be odd.
            if surface_dimensions is None: the time surface is defined globally, on the whole sensor grid.
        tau (float): time constant to decay events around occuring event with.
        decay (str): can be either 'lin' or 'exp', corresponding to linear or exponential decay.

    Returns:
        array of timesurfaces with dimensions (w,h) or (p,w,h)
    """

    if surface_dimensions:
        assert len(surface_dimensions) == 2
        assert surface_dimensions[0] % 2 == 1 and surface_dimensions[1] % 2 == 1
        radius_x = surface_dimensions[0] // 2
        radius_y = surface_dimensions[1] // 2
    else:
        radius_x = 0
        radius_y = 0
        surface_dimensions = sensor_size

    assert "x" and "y" and "t" and "p" in events.dtype.names

    timestamp_memory = np.zeros(
        (sensor_size[2], sensor_size[1] + radius_y * 2, sensor_size[0] + radius_x * 2)
    )
    timestamp_memory -= tau * 3 + 1
    all_surfaces = np.zeros(
        (len(events), sensor_size[2], surface_dimensions[1], surface_dimensions[0])
    )
    for index, event in enumerate(events):
        x = int(event["x"])
        y = int(event["y"])
        timestamp_memory[int(event["p"]), y + radius_y, x + radius_x] = event["t"]
        if radius_x > 0 and radius_y > 0:
            timestamp_context = (
                timestamp_memory[
                    :, y : y + surface_dimensions[1], x : x + surface_dimensions[0]
                ]
                - event["t"]
            )
        else:
            timestamp_context = timestamp_memory - event["t"]

        if decay == "lin":
            timesurface = timestamp_context / (3 * tau) + 1
            timesurface[timesurface < 0] = 0
        elif decay == "exp":
            timesurface = np.exp(timestamp_context / tau)
        all_surfaces[index, :, :, :] = timesurface
    return all_surfaces



import os
import shutil
import h5py
import numpy as np
from pathlib import Path
from tonic import datasets, transforms
from tonic import MemoryCachedDataset, DiskCachedDataset


def test_memory_caching_pokerdvs():
    dataset = datasets.POKERDVS(save_to="./data", train=False)
    cached_dataset = MemoryCachedDataset(dataset)
    assert len(dataset) == len(cached_dataset)

    # cache all samples
    for data, label in cached_dataset:
        pass

    assert len(cached_dataset.data_dict) == len(dataset)


def test_caching_pokerdvs():
    dataset = datasets.POKERDVS(save_to="./data", train=False)
    cache_path = "./cache/test1"
    cached_dataset = DiskCachedDataset(dataset, cache_path)

    # cache all samples
    for data, label in cached_dataset:
        pass

    assert len(dataset) == len(cached_dataset)
    assert len(os.listdir(cache_path)) == len(dataset)


def test_caching_transforms():
    sensor_size = datasets.POKERDVS.sensor_size
    preprocess = transforms.Compose(
        [transforms.Downsample(time_factor=1, spatial_factor=1)]
    )
    augmentation = transforms.Compose(
        [transforms.Downsample(time_factor=1, spatial_factor=1)]
    )
    dataset = datasets.POKERDVS(save_to="./data", train=True, transform=preprocess)

    dataset_cached = DiskCachedDataset(
        dataset, cache_path="./cache/test2", transform=augmentation, num_copies=4
    )

    for (data, label), (data2, label2) in zip(dataset, dataset_cached):
        assert (data == data2).all()
        assert label == label2


def test_caching_reset():
    cache_path = Path("./cache/test3")
    if cache_path.exists():
        shutil.rmtree(cache_path)

    # simulate outdated cache file
    old_file = cache_path / "0_0.hdf5"
    dummy_content = np.zeros((3, 3))
    os.makedirs(cache_path)
    with h5py.File(old_file, "w") as f:
        f.create_dataset(f"data/0", data=dummy_content)
        f.create_dataset(f"target/0", data=dummy_content)

    dataset = datasets.POKERDVS(save_to="./data", train=False)

    # load first sample from cache, which will load the old file
    dataset_cached = DiskCachedDataset(
        dataset, cache_path=cache_path, reset_cache=False
    )
    data, target = dataset_cached[0]
    assert (data == dummy_content).all()

    # when using reset_cache=True, the folder will be reset and the right file cached
    dataset_cached = DiskCachedDataset(dataset, cache_path=cache_path, reset_cache=True)
    data, target = dataset_cached[0]
    assert data.shape != dummy_content.shape
    assert data.shape[0] > 100


def test_caching_from_files():
    dataset = datasets.POKERDVS(save_to="./data", train=False)
    cache_path = "./cache/test4"
    dataset = DiskCachedDataset(dataset, cache_path)

    n_cached_samples = 10
    for i in range(n_cached_samples):
        events, target = dataset[i]

    dataset = DiskCachedDataset(dataset=None, cache_path=cache_path)
    assert len(dataset) == n_cached_samples
    assert len(os.listdir(cache_path)) == len(dataset)



import unittest
from typing import Dict, Any, Union, Optional
import numpy as np
from unittest.mock import patch
import shutil


class DatasetTestCase(unittest.TestCase):
    DATASET_CLASS = None
    FEATURE_TYPES = None

    _CHECK_FUNCTIONS = {"check_md5", "check_integrity", "check_exists"}
    _DOWNLOAD_EXTRACT_FUNCTIONS = {
        "download_url",
        "download_file_from_google_drive",
        "extract_archive",
        "download_and_extract_archive",
    }

    def inject_fake_data(
        self, tmpdir: str, config: Dict[str, Any]
    ) -> Union[int, Dict[str, Any]]:
        """Inject fake data for dataset into a temporary directory.
        During the creation of the dataset the download and extract logic is disabled. Thus, the fake data injected
        here needs to resemble the raw data, i.e. the state of the dataset directly after the files are downloaded and
        potentially extracted.
        Args:
            tmpdir (str): Path to a temporary directory. For most cases this acts as root directory for the dataset
                to be created and in turn also for the fake data injected here.
            config (Dict[str, Any]): Configuration that will be passed to the dataset constructor. It provides at least
                fields for all dataset parameters with default values.
        Needs to return one of the following:
            1. (int): Number of examples in the dataset to be created, or
            2. (Dict[str, Any]): Additional information about the injected fake data. Must contain the field
                ``"num_examples"`` that corresponds to the number of examples in the dataset to be created.
        """
        raise NotImplementedError(
            "You need to provide fake data in order for the tests to run."
        )

    def create_dataset(self, inject_fake_data: bool = True, **kwargs: Any):
        tmpdir = "/tmp/"
        info = self._inject_fake_data(tmpdir)

        if inject_fake_data:
            with patch.object(self.DATASET_CLASS, "_check_exists", return_value=True):
                dataset = self.DATASET_CLASS(**self.KWARGS)
        else:
            dataset = self.DATASET_CLASS(**self.KWARGS)
        return dataset, info

    def _inject_fake_data(self, tmpdir):
        info = self.inject_fake_data(tmpdir)
        if info is None:
            raise UsageError(
                "The method 'inject_fake_data' needs to return at least an integer indicating the number of "
                "examples for the current configuration."
            )
        return info

    def _patch_checks(self):
        return {
            patch.object(self.DATASET_CLASS, function, return_value=True)
            for function in self._CHECK_FUNCTIONS
        }

    #     def test_download_started(self):
    #         with pytest.raises((FileNotFoundError, RuntimeError)):
    #             dataset, info = self.create_dataset(inject_fake_data=False)

    def test_feature_types(self):
        dataset, info = self.create_dataset()
        data, target = dataset[0]

        if len(self.FEATURE_TYPES) == 1:
            data = (data,)
        if len(self.TARGET_TYPES) == 1:
            target = (target,)
        assert len(data) == len(self.FEATURE_TYPES)
        assert len(target) == len(self.TARGET_TYPES)

        for (data_piece, feature_type) in zip(data, self.FEATURE_TYPES):
            if type(data_piece) == np.ndarray:
                assert data_piece.dtype == feature_type
            else:
                assert type(data_piece) == feature_type

    def test_num_examples(self):
        dataset, info = self.create_dataset()
        assert len(dataset) == info["n_samples"]

    @classmethod
    def setUpClass(cls):
        cls.KWARGS.update({"save_to": "/tmp"})
        shutil.rmtree("/tmp/" + cls.DATASET_CLASS.__name__, ignore_errors=True)
        super().setUpClass()



from torchdata.datapipes.iter import (
    FileLister,
    Filter,
    Forker,
    Zipper,
)
from functools import partial
import os
import numpy as np
from pathlib import Path
from tonic.io import read_mnist_file
from tonic.download_utils import check_integrity, download_and_extract_archive

#####
# Dataset properties.
#####

sensor_size = (34, 34, 2)
dtype = np.dtype([("x", int), ("y", int), ("t", int), ("p", int)])
BASE_URL = "https://data.mendeley.com/public-files/datasets/468j46mzdv/files/"
TRAIN_URL = BASE_URL + "39c25547-014b-4137-a934-9d29fa53c7a0/file_downloaded"
TRAIN_FILENAME = "train.zip"
TRAIN_MD5 = "20959b8e626244a1b502305a9e6e2031"
TRAIN_FOLDER = "train"
TEST_URL = BASE_URL + "05a4d654-7e03-4c15-bdfa-9bb2bcbea494/file_downloaded"
TEST_FILENAME = "test.zip"
TEST_MD5 = "69ca8762b2fe404d9b9bad1103e97832"
TEST_FOLDER = "test"

#####
# Functions
#####


def _is_bin_file(data):
    return data.endswith("bin")


def _read_label_from_filepath(filepath):
    return int(filepath.split("/")[-2])


def _at_least_n_files(root, n_files, file_type):
    check = n_files <= len(list(Path(root).glob(f"**/*{file_type}")))
    return check


def _check_exists(filepath):
    return _at_least_n_files(filepath, n_files=100, file_type=".bin")


def _first_saccade_filter(events):
    return events[events["t"] < 1e5]


#####
# Dataset
#####


def nmnist(
    root, transform=None, target_transform=None, train=True, first_saccade_only=False
):
    # Setting file path depending on train value.
    filepath = root + "/" + (TRAIN_FOLDER if train else TEST_FOLDER)
    url = TRAIN_URL if train else TEST_URL
    md5 = TRAIN_MD5 if train else TEST_MD5
    filename = TRAIN_FILENAME if train else TEST_FILENAME
    # Downloading the MNIST file if it exists.
    if not _check_exists(filepath):
        download_and_extract_archive(
            url=url, 
            download_root=filepath, 
            filename=filename,
            md5=md5
        )
    # Creating the datapipe.
    dp = FileLister(root=filepath, recursive=True)
    dp = Filter(dp, _is_bin_file)
    # Thinking about avoiding this fork in order to apply transform to both targe and events.
    event_dp, label_dp = Forker(dp, num_instances=2)
    event_dp = event_dp.map(partial(read_mnist_file, dtype=dtype))
    if first_saccade_only:
        event_dp = event_dp.map(_first_saccade_filter)
    label_dp = label_dp.map(_read_label_from_filepath)
    if transform is not None:
        event_dp = event_dp.map(transform)
    if target_transform is not None:
        label_dp = label_dp.map(target_transform)
    dp = Zipper(event_dp, label_dp)
    return dp



"""
=============
ToTimesurface
=============

:class:`~tonic.transforms.ToTimesurface` creates local or global time surfaces for each event.
"""

import tonic

nmnist = tonic.datasets.NMNIST("../../tutorials/data", train=False)
events, label = nmnist[0]


####################################
# ToTimesurface global
# --------------------------------

transform = tonic.transforms.ToTimesurface(
    sensor_size=nmnist.sensor_size, surface_dimensions=None, tau=100000, decay="exp"
)

frames = transform(events)

# only plot a few of them
ani = tonic.utils.plot_animation(frames[::70])


####################################
# ToTimesurface local
# --------------------------------

transform = tonic.transforms.ToTimesurface(
    sensor_size=nmnist.sensor_size, surface_dimensions=(9, 9), tau=100000, decay="exp"
)

frames = transform(events)

# only plot a few of them
ani = tonic.utils.plot_animation(frames[::70])



import os
from typing import Callable, Optional

import numpy as np
from tonic.dataset import Dataset
from tonic.download_utils import extract_archive
from tonic.io import read_aedat4


class CIFAR10DVS(Dataset):
    """`CIFAR10-DVS <https://www.frontiersin.org/articles/10.3389/fnins.2017.00309/full>`_
    ::

        @article{li2017cifar10,
        title={Cifar10-dvs: an event-stream dataset for object classification},
        author={Li, Hongmin and Liu, Hanchao and Ji, Xiangyang and Li, Guoqi and Shi, Luping},
        journal={Frontiers in neuroscience},
        volume={11},
        pages={309},
        year={2017},
        publisher={Frontiers}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    url = "http://cifar10dvs.ridger.top/CIFAR10DVS.zip"

    filename = "CIFAR10DVS.zip"
    file_md5 = "ce3a4a0682dc0943703bd8f749a7701c"
    data_filename = [
        "airplane.zip",
        "automobile.zip",
        "bird.zip",
        "cat.zip",
        "deer.zip",
        "dog.zip",
        "frog.zip",
        "horse.zip",
        "ship.zip",
        "truck.zip",
    ]

    folder_name = "CIFAR10DVS"
    dtype = np.dtype(
        [("t", np.uint64), ("x", np.uint16), ("y", np.uint16), ("p", bool)]
    )
    ordering = dtype.names
    sensor_size = (128, 128, 2)

    def __init__(
        self,
        save_to: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )

        # classes for CIFAR10DVS dataset

        classes = {
            "airplane": 0,
            "automobile": 1,
            "bird": 2,
            "cat": 3,
            "deer": 4,
            "dog": 5,
            "frog": 6,
            "horse": 7,
            "ship": 8,
            "truck": 9,
        }

        if not self._check_exists():
            self.download()
            for filename in self.data_filename:
                extract_archive(os.path.join(self.location_on_system, filename))

        file_path = os.path.join(self.location_on_system, self.folder_name)
        for path, dirs, files in os.walk(file_path):
            dirs.sort()
            for file in files:
                if file.endswith("aedat4"):
                    self.data.append(path + "/" + file)
                    label_number = classes[os.path.basename(path)]
                    self.targets.append(label_number)

    def __getitem__(self, index):
        """
        Returns:
            a tuple of (events, target) where target is the index of the target class.
        """
        events = read_aedat4(self.data[index])
        events.dtype.names = ["t", "x", "y", "p"]

        target = self.targets[index]

        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.transforms is not None:
            events, target = self.transforms(events, target)
        return events, target

    def __len__(self):
        return len(self.data)

    def _check_exists(self):
        return (
            self._is_file_present()
            and self._folder_contains_at_least_n_files_of_type(1000, ".aedat4")
        )



from .asl_dvs import ASLDVS
from .cifar10dvs import CIFAR10DVS
from .davisdataset import DAVISDATA
from .dsec import DSEC
from .dvsgesture import DVSGesture
from .hsd import SHD, SSC
from .mvsec import MVSEC
from .ncaltech101 import NCALTECH101
from .nmnist import NMNIST
from .pokerdvs import POKERDVS
from .s_mnist import SMNIST
from .tum_vie import TUMVIE
from .visual_place_recognition import VPR
from .dvs_lips import DVSLip

__all__ = [
    "ASLDVS",
    "CIFAR10DVS",
    "DAVISDATA",
    "DSEC",
    "DVSGesture",
    "MVSEC",
    "NCALTECH101",
    "NMNIST",
    "POKERDVS",
    "SHD",
    "SMNIST",
    "SSC",
    "TUMVIE",
    "VPR",
    "DVSLip",
]



import os
from typing import Callable, Optional

import numpy as np
from tonic.dataset import Dataset


class DVSGesture(Dataset):
    """`IBM DVS Gestures <http://research.ibm.com/dvsgesture/>`_

    .. note::  This is (exceptionally) a preprocessed version of the original dataset,
               where recordings that originally contained multiple labels have already
               been cut into respective samples. Also temporal precision is reduced to ms.

    ::

        @inproceedings{amir2017low,
          title={A low power, fully event-based gesture recognition system},
          author={Amir, Arnon and Taba, Brian and Berg, David and Melano, Timothy and McKinstry, Jeffrey and Di Nolfo, Carmelo and Nayak, Tapan and Andreopoulos, Alexander and Garreau, Guillaume and Mendoza, Marcela and others},
          booktitle={Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition},
          pages={7243--7252},
          year={2017}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        train (bool): If True, uses training subset, otherwise testing subset.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    # Train: https://www.neuromorphic-vision.com/public/downloads/ibmGestureTrain.tar.gz
    # Test : https://www.neuromorphic-vision.com/public/downloads/ibmGestureTest.tar.gz
    base_url = "https://www.neuromorphic-vision.com/public/downloads/"
    test_url = base_url + "ibmGestureTest.tar.gz"
    train_url = base_url + "ibmGestureTrain.tar.gz"
    test_md5 = "56070e45dadaa85fff82e0fbfbc06de5"
    train_md5 = "3a8f0d4120a166bac7591f77409cb105"
    test_filename = "ibmGestureTest.tar.gz"
    train_filename = "ibmGestureTrain.tar.gz"
    classes = [
        "hand_clapping",
        "right_hand_wave",
        "left_hand_wave",
        "right_arm_clockwise",
        "right_arm_counter_clockwise",
        "left_arm_clockwise",
        "left_arm_counter_clockwise",
        "arm_roll",
        "air_drums",
        "air_guitar",
        "other_gestures",
    ]

    sensor_size = (128, 128, 2)
    dtype = np.dtype([("x", np.int16), ("y", np.int16), ("p", bool), ("t", np.int64)])
    ordering = dtype.names

    def __init__(
        self,
        save_to: str,
        train: bool = True,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )
        self.train = train

        if train:
            self.url = self.train_url
            self.file_md5 = self.train_md5
            self.filename = self.train_filename
            self.folder_name = "ibmGestureTrain"
        else:
            self.url = self.test_url
            self.file_md5 = self.test_md5
            self.filename = self.test_filename
            self.folder_name = "ibmGestureTest"

        if not self._check_exists():
            self.download()

        file_path = os.path.join(self.location_on_system, self.folder_name)
        for path, dirs, files in os.walk(file_path):
            dirs.sort()
            for file in files:
                if file.endswith("npy"):
                    self.data.append(path + "/" + file)
                    self.targets.append(int(file[:-4]))

    def __getitem__(self, index):
        """
        Returns:
            a tuple of (events, target) where target is the index of the target class.
        """
        events = np.load(self.data[index])
        events[:, 3] *= 1000  # convert from ms to us
        events = np.lib.recfunctions.unstructured_to_structured(events, self.dtype)
        target = self.targets[index]
        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.transforms is not None:
            events, target = self.transforms(events, target)
        return events, target

    def __len__(self):
        return len(self.data)

    def _check_exists(self):
        return (
            self._is_file_present()
            and self._folder_contains_at_least_n_files_of_type(100, ".npy")
        )



#! /usr/bin/env python3

import numpy as np
import math


def _gen_pixel_to_cell_mapper(h, w, cell_size):
    """
    Matrix that allows to map a pixel to a cell via indexing.
    """
    matrix = np.zeros((h, w), dtype=np.int32)
    for i in range(h):
        for j in range(w):
            # Since each frame axis is divided in cells of width _K, the cell index related to that axis can be computed by integer division x//_K.
            cell_y, cell_x = i // cell_size, j // cell_size
            # Getting the index associated to cell coordinates.
            matrix[i, j] = cell_y * math.ceil(w / cell_size) + cell_x
    return matrix


def _get_cell_mems(events, n_cells, n_pols, pixel_to_cell_mapper):
    """
    Generates the local memories of all cells in the frame.
    """
    cell_mems = [[[] for c in range(n_cells)] for p in range(n_pols)]
    for event in events:
        cell_mems[int(event["p"])][pixel_to_cell_mapper[event["y"], event["x"]]].append(
            event
        )
    cell_mems = [
        [
            np.hstack(cell_mems[p][c]) if cell_mems[p][c] else np.empty((0,))
            for c in range(n_cells)
        ]
        for p in range(n_pols)
    ]
    return cell_mems


def _get_time_surface(
    event, cell_mem, surface_size=10, temporal_window=500, tau=500, decay="exp"
):
    """
    Accumulates the time surfaces of all the events included in the time window of the given event.
    """
    # Empty cell.
    if cell_mem.size == 0:
        return np.zeros((surface_size, surface_size))
    radius = np.array(surface_size // 2).astype(cell_mem["x"].dtype)
    # Getting starting timestamp.
    t_start = max(0, event["t"] - temporal_window)
    # Computing surface coordinates for events.
    x_surface, y_surface = (event["x"] + radius - cell_mem["x"]), (
        event["y"] + radius - cell_mem["y"]
    )
    # Filtering events out of surface and time window.
    mask = np.asarray(
        (x_surface >= 0)
        & (x_surface <= 2 * radius)
        & (y_surface >= 0)
        & (y_surface <= 2 * radius)
        & (cell_mem["t"] >= t_start)
    ).nonzero()[0]
    if mask.size == 0:
        return np.zeros((surface_size, surface_size))
    # Time surfaces tensor.
    ts = np.zeros((mask.size, surface_size, surface_size), dtype=np.float32)
    # For each event, the corresponding time surface is computed using the coordinates previously calculated.
    ts[np.arange(mask.size), y_surface[mask], x_surface[mask]] = (
        np.exp(-(event["t"] - cell_mem["t"][mask]).astype(np.float32) / tau)
        if decay == "exp"
        else (event["t"] - cell_mem["t"][mask]).astype(np.float32) / (3 * tau) + 1
    )
    # The time surfaces are accumulated and returned in output.
    return np.sum(ts, axis=0)


def to_averaged_timesurface(
    events,
    sensor_size=(128, 128, 2),
    cell_size=10,
    surface_size=3,
    temporal_window=5e5,
    tau=5e3,
    decay="exp",
    num_workers=1,
):
    """Representation that creates averaged timesurfaces for each event for one recording. Taken from the paper
    Sironi et al. 2018, HATS: Histograms of averaged time surfaces for robust event-based object classification
    https://openaccess.thecvf.com/content_cvpr_2018/papers/Sironi_HATS_Histograms_of_CVPR_2018_paper.pdf
    Parameters:
        cell_size (int): size of each square in the grid
        surface_size (int): has to be odd
        temporal_window (int): how far back to look for past events for the time averaging. Expressed in microseconds.
        tau (int): time constant to decay events around occuring event with. Expressed in microseconds.
        decay (str): can be either 'lin' or 'exp', corresponding to linear or exponential decay.
        merge_polarities (bool): flag that tells whether polarities should be taken into account separately or not.
        num_workers (int): number of workers to be deployed on the histograms computation.
    Returns:
        array of histograms (numpy.Array with shape (n_cells, n_pols, surface_size, surface_size))
    """

    assert surface_size <= cell_size
    assert surface_size % 2 != 0
    assert "x" and "y" and "t" and "p" in events.dtype.names
    assert decay == "lin" or decay == "exp"

    if num_workers > 1:
        try:
            from joblib import Parallel, delayed
        except Exception as e:
            print("Error: num_workers>1 needs joblib installed.")
        use_joblib = True
    else:
        use_joblib = False

    # Getting sensor sizes, number of cells in the frame and initializing the data structures.
    width, height, n_pols = sensor_size
    n_cells = math.ceil(width / cell_size) * math.ceil(height / cell_size)
    histograms = np.zeros((n_cells, n_pols, surface_size, surface_size))

    # Matrix for associating an event to a cell via indexing.
    pixel_to_cell_mapper = _gen_pixel_to_cell_mapper(
        w=width, h=height, cell_size=cell_size
    )

    # Organizing the events in cells (local memories of HATS paper).
    cell_mems = _get_cell_mems(
        events=events,
        n_cells=n_cells,
        n_pols=n_pols,
        pixel_to_cell_mapper=pixel_to_cell_mapper,
    )

    # Time surfaces associated to each event in a cell.
    get_cell_time_surfaces = lambda cell_mem: np.stack(
        [
            _get_time_surface(
                event=cell_mem[i],
                cell_mem=cell_mem[:i],
                surface_size=surface_size,
                temporal_window=temporal_window,
                tau=tau,
                decay=decay,
            )
            for i in range(cell_mem.shape[0])
        ]
    )

    # Histogram associated to a cell, obtained by accumulatig the time surfaces of each event in the cell.
    get_cell_histogram = lambda cell: np.stack(
        [
            np.sum(get_cell_time_surfaces(cell_mems[pol][cell]), axis=0)
            / max(sum([cell_mems[p][cell].shape[0] for p in range(n_pols)]), 1)
            if cell_mems[pol][cell].size > 0
            else np.zeros((surface_size, surface_size))
            for pol in range(n_pols)
        ]
    )

    if not use_joblib:
        histograms = np.stack([get_cell_histogram(c) for c in range(n_cells)])
    else:
        histograms = np.stack(
            Parallel(n_jobs=num_workers)(
                delayed(get_cell_histogram)(c) for c in range(n_cells)
            )
        )
    return histograms



import os
from typing import Any, Callable, Optional, Tuple

import numpy as np
import scipy.io as scio
from tonic.dataset import Dataset
from tonic.download_utils import extract_archive
from tonic.io import make_structured_array


class ASLDVS(Dataset):
    """`ASL-DVS <https://github.com/PIX2NVS/NVS2Graph>`_
    ::

        @inproceedings{bi2019graph,
            title={Graph-based Object Classification for Neuromorphic Vision Sensing},
            author={Bi, Y and Chadha, A and Abbas, A and and Bourtsoulatze, E and Andreopoulos, Y},
            booktitle={2019 IEEE International Conference on Computer Vision (ICCV)},
            year={2019},
            organization={IEEE}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    url = "https://www.dropbox.com/sh/ibq0jsicatn7l6r/AACNrNELV56rs1YInMWUs9CAa?dl=1"
    filename = "ASLDVS.zip"
    file_md5 = "33f8b87bf45edc0bfed0de41822279b9"
    folder_name = ""

    classes = [chr(letter) for letter in range(97, 123)]  # generate alphabet
    int_classes = dict(zip(classes, range(len(classes))))
    sensor_size = (240, 180, 2)
    dtype = np.dtype([("t", int), ("x", int), ("y", int), ("p", int)])
    ordering = dtype.names

    def __init__(
        self,
        save_to: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )

        if not self._check_exists():
            self.download()
            # extract zips within zip
            for path, dirs, files in os.walk(self.location_on_system):
                dirs.sort()
                for file in files:
                    if file.startswith("Yin") and file.endswith("zip"):
                        extract_archive(os.path.join(self.location_on_system, file))

        for path, dirs, files in os.walk(self.location_on_system):
            dirs.sort()
            files.sort()
            for file in files:
                if file.endswith("mat"):
                    self.data.append(path + "/" + file)
                    self.targets.append(self.int_classes[path[-1]])

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """
        Returns:
            (events, target) where target is index of the target class.
        """
        events, target = scio.loadmat(self.data[index]), self.targets[index]
        events = make_structured_array(
            events["ts"],
            events["x"],
            self.sensor_size[1] - 1 - events["y"],
            events["pol"],
            dtype=self.dtype,
        )
        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.transforms is not None:
            events, target = self.transforms(events, target)
        return events, target

    def __len__(self):
        return len(self.data)

    def _check_exists(self):
        return (
            self._is_file_present()
            and self._folder_contains_at_least_n_files_of_type(100800, ".mat")
        )



"""
==========
CenterCrop
==========
:class:`~tonic.transforms.CenterCrop` crops the events to a central size.
"""

import tonic

nmnist = tonic.datasets.NMNIST("../../tutorials/data", train=False)
events, label = nmnist[0]

cropped_size = (18, 18)
transform = tonic.transforms.Compose(
    [
        tonic.transforms.CenterCrop(sensor_size=nmnist.sensor_size, size=cropped_size),
        tonic.transforms.ToFrame(
            sensor_size=(*cropped_size, 2),
            time_window=10000,
        ),
    ]
)

frames = transform(events)

ani = tonic.utils.plot_animation(frames)



import librosa
import numpy as np
from dataclasses import dataclass
from scipy.signal import butter, sosfilt
from typing import Tuple, List, Union, Iterator


@dataclass
class FixLength:
    """
    Fix the length of a sample along a specified axis to a given length.

    Parameters:
        length:
            Desired length of the sample
        axis:
            Dimension along which the length needs to be fixed.
    Args:
        data: torch.Tensor
    Returns:
        torch.Tensor of the same dimension
    """

    length: int
    axis: int = 1

    def __call__(self, data: np.ndarray):
        return librosa.util.fix_length(data, self.length, self.axis)


@dataclass
class Bin:
    """
    Bin the given data along a specified axis at the specified new frequency

    Parameters:
        orig_freq: float
            Sampling frequency of the given data stream
        new_freq: float
            Desired frequency after binning
        axis: int
            Axis along which the data needs to be binned

    Args:
         data: torch.Tensor
            data to be binned

    Returns:
        torch.Tensor binned data

    """

    orig_freq: float
    new_freq: float
    axis: int

    def __call__(self, data: np.ndarray):
        data_len = data.shape[self.axis]
        n_splits = int(data_len / (self.orig_freq / self.new_freq))
        splits = np.array_split(data, n_splits, axis=self.axis)
        data = [np.sum(split, axis=self.axis, keepdims=True) for split in splits]
        return np.concatenate(data, self.axis)


@dataclass
class SOSFilter:
    """
    SOS filter

    Parameters
    ----------

    coeffs:
        coefficients of the second order filter
    axis:
        Axis along with the filter needs to be applied

    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.sosfilt.html for more details
    """

    coeffs: np.ndarray
    axis: int

    def __call__(self, signal):
        return sosfilt(self.coeffs, signal, axis=self.axis)


@dataclass
class ButterFilter:
    """
    Butter filter

    Parameters
    ----------

    order:
        Order of filter to be used
    freq:
        Frequency for the filter (float or (float, float))
    analog:
        True if analog filter
    btype:
        Filter type, {‘lowpass’, ‘highpass’, ‘bandpass’, ‘bandstop’}
    rectify:
        If true, the output is the absolute value of the filtered output
    axis:
        Axis along which the filter needs to be applied

    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.butter.html#scipy.signal.butter for more details on parameters.
    """

    order: int
    freq: Union[float, Tuple[float, float]]
    analog: bool
    btype: str
    rectify: bool
    axis: int

    def __post_init__(self):
        coeffs = butter(
            self.order, self.freq, analog=self.analog, btype=self.btype, output="sos"
        )
        self.filter = SOSFilter(coeffs, axis=self.axis)

    def __call__(self, data: np.ndarray) -> np.ndarray:
        out = self.filter(data)
        if self.rectify:
            return np.abs(out)
        else:
            return out


@dataclass
class ButterFilterBank:
    """
    Butter filter bank

    Parameters
    ----------

    order:
        Order of filter to be used
    freq:
        Frequency for the filter (float or (float, float))
    rectify:
        If true, the output is the absolute value of the filtered output
    axis:
        Axis along which the filter needs to be applied
    analog:
        If true, the filter will be analog. False by default

    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.butter.html#scipy.signal.butter for more details on parameters.
    """

    order: int
    freq: List[Tuple[float, float]]
    rectify: bool
    axis: int
    analog: bool = False

    def __post_init__(self):
        self.filters = [
            ButterFilter(
                self.order,
                freq,
                analog=self.analog,
                btype="band",
                rectify=self.rectify,
                axis=self.axis,
            )
            for freq in self.freq
        ]

    def __call__(self, data):
        return np.concatenate([filt(data) for filt in self.filters], axis=0)


@dataclass
class LinearButterFilterBank:
    """
    Butter filter bank

    Parameters
    ----------

    order:
        Order of filter to be used
    low_freq:
        Lower/cutoff frequency the filter (float or (float, float))
    sampling_freq:
        Sampling frequency of the signal, also serves as higher frequency of the filter bank.
    analog:
        True if analog filter
    rectify:
        If true, the output is the absolute value of the filtered output
    axis:
        Axis along which the filter needs to be applied

    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.butter.html#scipy.signal.butter for more details on parameters.
    """

    order: int = 2
    low_freq: float = 100
    sampling_freq: float = 16000
    analog: bool = False
    num_filters: int = 64
    rectify: bool = True
    axis: int = -1

    def compute_freq_bands(self):
        filter_bandwidth = 2 / self.num_filters
        nyquist = self.sampling_freq / 2

        high_freq = self.sampling_freq / 2 / (1 + filter_bandwidth) - 1
        freqs = np.linspace(self.low_freq, high_freq, self.num_filters)

        return np.array([freqs, freqs * (1 + filter_bandwidth)]).T / nyquist

    def __post_init__(self):
        freq_bands = self.compute_freq_bands()
        self.filterbank = ButterFilterBank(
            order=self.order, freq=freq_bands, rectify=self.rectify, axis=self.axis
        )

    def __call__(self, data):
        return self.filterbank(data)


@dataclass
class MelButterFilterBank(LinearButterFilterBank):
    """
    Butter filter bank with frequencies along the mel scale

    Parameters
    ----------

    order:
        Order of filter to be used
    low_freq:
        Lower/cutoff frequency the filter (float or (float, float))
    sampling_freq:
        Sampling frequency of the signal, also serves as higher frequency of the filter bank.
    analog:
        True if analog filter
    rectify:
        If true, the output is the absolute value of the filtered output
    axis:
        Axis along which the filter needs to be applied

    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.butter.html#scipy.signal.butter for more details on parameters.
    """

    @staticmethod
    def hz2mel(freq):
        return 2595 * np.log10(1 + freq / 700)

    @staticmethod
    def mel2hz(freq):
        return 700 * (10 ** (freq / 2595) - 1)

    def compute_freq_bands(self):
        filter_bandwidth = 2 / self.num_filters
        nyquist = self.sampling_freq / 2

        high_freq = self.sampling_freq / 2 / (1 + filter_bandwidth) - 1
        freqs = np.linspace(self.low_freq, high_freq, self.num_filters)

        freq_bands = np.array([freqs, freqs * (1 + filter_bandwidth)]) / nyquist

        low_freq = self.hz2mel(self.low_freq)
        high_freq = self.hz2mel(self.sampling_freq / 2 / (1 + filter_bandwidth) - 1)
        freqs = self.mel2hz(np.linspace(low_freq, high_freq, self.num_filters))

        return np.array([freqs, freqs * (1 + filter_bandwidth)]).T / nyquist


@dataclass
class AddNoise:
    """
    Add nose to data

    Params:
        dataset:
            A dataset object that returns a tuple when iterated over the first element of which is the audio signal to be used for noise.
        snr:
            Desired signal to noise ratio in dB
        normed:
            If set to false, the signal max value will not be normalized. True by default.
    """

    dataset: Iterator
    snr: float
    normed: bool = True

    def get_noise_sample(self, sample_len: int) -> np.ndarray:
        """Get a random noise sample from the dataset"""
        # Find noise sample of minimum length
        while True:
            noise_idx = np.random.randint(0, len(self.dataset), (1,)).item()
            noise = self.dataset[noise_idx][0]
            if noise.shape[1] >= sample_len:
                break
        # Sample a random part of the data recording
        noise_signal_len = noise.shape[1]
        if noise_signal_len > sample_len:
            start_t = np.random.randint(0, noise_signal_len - sample_len, (1,)).item()
            noise = noise[:, start_t : start_t + sample_len]
        return noise

    def __call__(self, signal):
        # randomly pick a piece of noise data
        noise = self.get_noise_sample(sample_len=signal.shape[1])

        # mix signal with noise with given SNR
        signal_power = (signal**2).mean()
        noise_power = (noise**2).mean()
        noise_scale = (signal_power / noise_power) * 10 ** (-self.snr / 10)
        signal_with_snr = signal + noise_scale * noise

        # Normalize if specified
        if self.normed:
            return normalize(signal_with_snr)
        else:
            return signal_with_snr


def normalize(signal):
    """Normalize the signal"""
    signal -= signal.mean()
    max_val = np.max(np.abs(signal))
    if max_val > 0:
        signal /= max_val
    return signal


# @dataclass
# class DivisiveNormalization:
#    frame_dt: float  # Frame clock step
#    num_frames_avg: int  # Number of frames to average over
#    gating_clock_dt: float  # Clock frequency of gating E(t)
#    dt: float = 1  # Global clock step
#
#    def __call__(self, events: np.ndarray):
#        raise NotImplementedError



from . import datasets, transforms, collation, io, slicers, utils
from .dataset import Dataset
from .cached_dataset import CachedDataset, DiskCachedDataset, MemoryCachedDataset
from .sliced_dataset import SlicedDataset
from pbr.version import VersionInfo

all = "__version__"
__version__ = VersionInfo("tonic").release_string()



from typing import Tuple, Union
import numpy as np


def drop_event_numpy(events: np.ndarray, drop_probability: float):
    """Randomly drops events with drop_probability.

    Parameters:
        events: ndarray of shape [num_events, num_event_channels].
        drop_probability: probability of dropping out event.

    Returns:
        augmented events that were not dropped.
    """

    n_events = events.shape[0]
    n_dropped_events = int(drop_probability * n_events + 0.5)
    dropped_event_indices = np.random.choice(n_events, n_dropped_events, replace=False)
    return np.delete(events, dropped_event_indices, axis=0)


def drop_by_time_numpy(
    events: np.ndarray,
    duration_ratio: Union[float, Tuple[float]] = 0.2,
):
    """Drops events in a certain time interval with a length proportional to a specified ratio of the original length.

    Args:
        events (np.ndarray): ndarray of shape [num_events, num_event_channels]
        duration_ratio (Union[float, Tuple[float]], optional): the length of the dropped time interval, expressed in a ratio of the original sequence duration.
            - If a float, the value is used to calculate the interval length
            - If a tuple of 2 floats, the ratio is randomly chosen in [min, max).
            Defaults to 0.2.

    Returns:
        np.ndarray: augmented events that were not dropped (i.e., the events that are not in the time interval).
    """
    assert "x" and "t" and "p" in events.dtype.names
    assert (
        type(duration_ratio) == float and duration_ratio >= 0.0 and duration_ratio < 1.0
    ) or (
        type(duration_ratio) == tuple
        and len(duration_ratio) == 2
        and all(val >= 0 and val < 1.0 for val in duration_ratio)
    )

    # time interval
    t_start = 0.0
    t_end = events["t"].max()

    if type(duration_ratio) is tuple:
        duration_ratio = np.random.uniform(duration_ratio[0], duration_ratio[1])

    drop_duration = (t_end - t_start) * duration_ratio

    drop_start = np.random.uniform(t_start, t_end - drop_duration)
    mask_events = (events["t"] >= drop_start) & (
        events["t"] <= drop_start + drop_duration
    )

    return np.delete(events, mask_events)  # remove events


def drop_by_area_numpy(
    events: np.ndarray, sensor_size: Tuple, area_ratio: Union[float, Tuple[float]] = 0.2
):
    """Drops events located in a randomly chosen box area. The size of the box area is defined by a specified ratio of the sensor size.

    Args:
        events (np.ndarray): ndarray of shape [num_events, num_event_channels]
        sensor_size (Tuple): size of the sensor that was used [W,H,P]
        area_ratio (Union[float, Tuple[float]], optional): Ratio of the sensor resolution that determines the size of the box area where events are dropped.
            - if a float, the value is used to calculate the size of the box area
            - if a tuple of 2 floats, the ratio is randomly chosen in [min, max)
            Defaults to 0.2.

    Returns:
        np.ndarray: augmented events that were not dropped (i.e., the events that are not located in the box area).
    """
    assert "x" and "t" and "y" and "p" in events.dtype.names
    assert (type(area_ratio) == float and area_ratio >= 0.0 and area_ratio < 1.0) or (
        type(area_ratio) is tuple
        and len(area_ratio) == 2
        and all(val >= 0 and val < 1.0 for val in area_ratio)
    )

    if not sensor_size:
        sensor_size_x = int(events["x"].max() + 1)
        sensor_size_p = len(np.unique(events["p"]))
        sensor_size_y = int(events["y"].max() + 1)
        sensor_size = (sensor_size_x, sensor_size_y, sensor_size_p)

    # select ratio
    if type(area_ratio) is tuple:
        area_ratio = np.random.uniform(area_ratio[0] and area_ratio[1])

    # select area
    cut_w = int(sensor_size[0] * area_ratio)
    cut_h = int(sensor_size[1] * area_ratio)
    bbx1 = np.random.randint(0, (sensor_size[0] - cut_w))
    bby1 = np.random.randint(0, (sensor_size[1] - cut_h))
    bbx2 = bbx1 + cut_w - 1
    bby2 = bby1 + cut_h - 1

    # filter image
    mask_events = (
        (events["x"] >= bbx1)
        & (events["y"] >= bby1)
        & (events["x"] <= bbx2)
        & (events["y"] <= bby2)
    )

    # delete events of bbox
    return np.delete(events, mask_events)  # remove events



from dataclasses import dataclass
from . import functional
from typing import Callable, Optional, Tuple, Union
import numpy as np


class Compose:
    """Composes several transforms together. This a literal copy of torchvision.transforms.Compose function for convenience.

    Parameters:
        transforms (list of ``Transform`` objects): list of transform(s) to compose.
                                                    Can combine Tonic, PyTorch Vision/Audio transforms.

    Example:
        >>> transforms.Compose([
        >>>     transforms.Denoise(filter_time=10000),
        >>>     transforms.ToFrame(n_time_bins=3),
        >>> ])
    """

    def __init__(self, transforms: Callable):
        self.transforms = transforms

    def __call__(self, events):
        for t in self.transforms:
            events = t(events)
        return events

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"
        return format_string


@dataclass
class CenterCrop:
    """Crops events at the center to a specific output size. If output size
    is smaller than input sensor size along any dimension, padding will be used,
    which doesn't influence the number of events but just their spatial location
    after cropping.
    """

    sensor_size: Tuple[int, int, int]
    size: Union[int, Tuple[int, int]]

    def __call__(self, events: np.ndarray) -> np.ndarray:
        if type(self.size) == int:
            self.size = [self.size, self.size]
        offsets = (self.sensor_size[0] - self.size[0]) // 2, (
            self.sensor_size[1] - self.size[1]
        ) // 2
        offset_idx = [max(offset, 0) for offset in offsets]
        cropped_events = events[
            (offset_idx[0] <= events["x"])
            & (events["x"] < (offset_idx[0] + self.size[0]))
            & (offset_idx[1] <= events["y"])
            & (events["y"] < (offset_idx[1] + self.size[1]))
        ]
        cropped_events["x"] -= offsets[0]
        cropped_events["y"] -= offsets[1]
        return cropped_events


@dataclass
class CropTime:
    """Drops events with timestamps below min and above max.

    Parameters:
        min (int): The minimum timestamp below which all events are dropped. Zero by default.
        max (int): The maximum timestamp above which all events are dropped.

    Example:
        >>> transform = tonic.transforms.CropTime(min=1000, max=20000)
    """

    min: int = 0
    max: int = None

    def __call__(self, events):
        assert "t" in events.dtype.names
        if self.max is None:
            self.max = np.max(events["t"])
        return events[(events["t"] >= self.min) & (events["t"] <= self.max)]


@dataclass(frozen=True)
class Denoise:
    """Drops events that are spatio-temporally not sufficiently close enough
    to other events in the sample. In practise that means that an event is
    dropped if no other event occured within a spatial neighbourhood
    of 1 pixel and a temporal neighbourhood of filter_time time units.
    Useful to filter noisy recordings where events occur isolated in time.

    Parameters:
        filter_time (float): minimum temporal distance to next event, otherwise dropped.
                    Lower values will mean higher constraints, therefore less output events.
                    Use same unit of time as the events have.

    Example:
        >>> transform = tonic.transforms.Denoise(filter_time=10000)
    """

    filter_time: float

    def __call__(self, events):
        return functional.denoise_numpy(events=events, filter_time=self.filter_time)


@dataclass
class Decimation:
    """Deterministically drops every nth event for every spatial location x (and potentially y).

    Parameters:
        n (int): The event stream for each x/y location is reduced to 1/n.

    Example:
        >>> transform = tonic.transforms.Decimation(n=5)
    """

    n: int

    def __call__(self, events):
        return functional.decimate_numpy(events=events, n=self.n)


@dataclass(frozen=True)
class DropEvent:
    """Randomly drops events with probability p. If random_p is selected, the drop probability is randomized between 0 and p.

    Parameters:
        p (float or tuple of floats): Probability of dropping events. Can be a tuple of floats (p_min, p_max), so that p is sampled from the range.

    Example:
        >>> transform1 = tonic.transforms.DropEvent(p=0.2)
        >>> transform2 = tonic.transforms.DropEvent(p=(0, 0.5))
    """

    p: Union[float, Tuple[float, float]]

    @staticmethod
    def get_params(p: Union[float, Tuple[float, float]]):
        if type(p) == tuple:
            p = (p[1] - p[0]) * np.random.random_sample() + p[0]
        return p

    def __call__(self, events):
        p = self.get_params(p=self.p)
        return functional.drop_event_numpy(events=events, drop_probability=p)


@dataclass(frozen=True)
class DropEventByTime:
    """Drops events in a certain time interval with a length proportional to a specified ratio of the original length.

    Parameters:
        duration_ratio (Union[float, Tuple[float]], optional): the length of the dropped time interval, expressed in a ratio of the original sequence duration.
            - If a float, the value is used to calculate the interval length
            - If a tuple of 2 floats, the ratio is randomly chosen in [min, max)
            Defaults to 0.2.

    Example:
        >>> transform = tonic.transforms.DropEventByTime(duration_ratio=(0.1, 0.8))
    """

    duration_ratio: Union[float, Tuple[float]] = 0.2

    def __call__(self, events):

        return functional.drop_by_time_numpy(events, self.duration_ratio)


@dataclass(frozen=True)
class DropEventByArea:
    """Drops events located in a randomly chosen box area. The size of the box area is defined by a specified ratio of the sensor size.

    Args:
        sensor_size (Tuple): size of the sensor that was used [W,H,P]
        area_ratio (Union[float, Tuple[float]], optional): Ratio of the sensor resolution that determines the size of the box area where events are dropped.
            - if a float, the value is used to calculate the size of the box area
            - if a tuple of 2 floats, the ratio is randomly chosen in [min, max)
            Defaults to 0.2.

    Example:
        >>> transform = tonic.transforms.DropEventByArea(sensor_size=(128,128,2), area_ratio=(0.1, 0.8))
    """

    sensor_size: Tuple[int, int, int]
    area_ratio: Union[float, Tuple[float]] = 0.2

    def __call__(self, events):

        return functional.drop_by_area_numpy(events, self.sensor_size, self.area_ratio)


@dataclass(frozen=True)
class EventDrop:
    """Applies EventDrop transformation from the paper "EventDrop: Data Augmentation for Event-based Learning".
        Applies one of the 4 drops of event strategies between:
            1. Identity (do nothing)
            2. Drop events by time
            3. Drop events by area
            4. Drop events randomly

        For each strategy, the ratio of dropped events are determined in the paper.

    Args:
        sensor_size (Tuple): size of the sensor that was used [W,H,P]

    Example:
        >>> transform = tonic.transforms.EventDrop(sensor_size=(128,128,2))
    """

    sensor_size: Tuple[int, int, int]

    def __call__(self, events):
        choice = np.random.randint(0, 4)
        if choice == 0:
            return events
        if choice == 1:
            duration_ratio = np.random.randint(1, 10) / 10.0
            return functional.drop_by_time_numpy(events, duration_ratio)
        if choice == 2:
            area_ratio = np.random.randint(1, 6) / 20.0
            return functional.drop_by_area_numpy(events, self.sensor_size, area_ratio)
        if choice == 3:
            ratio = np.random.randint(1, 10) / 10.0
            return functional.drop_event_numpy(events, ratio)


@dataclass
class DropPixel:
    """Drops events for individual pixels. If the locations of pixels to be dropped is known, a
    list of x/y coordinates can be passed directly. Alternatively, a cutoff frequency for each pixel can be defined
    above which pixels will be deactivated completely. This prevents so-called *hot pixels* which fire constantly
    (e.g. due to faulty hardware).

    Parameters:
        coordinates: list of (x,y) coordinates for which all events will be deleted.
        hot_pixel_frequency: drop pixels completely that fire higher than the given frequency.

    Example:
        >>> from tonic.transforms import DropPixel
        >>> transform1 = DropPixel(coordinates=[[10,10], [10,11], [11,10], [11,11]])
        >>> transform2 = DropPixel(hot_pixel_frequency=60) # Hertz
    """

    coordinates: Optional[Tuple] = None
    hot_pixel_frequency: Optional[int] = None

    def __call__(self, events):

        if events.dtype.names is not None:
            # assert "x", "y", "p" in events.dtype.names
            if self.hot_pixel_frequency:
                self.coordinates = functional.identify_hot_pixel(
                    events=events, hot_pixel_frequency=self.hot_pixel_frequency
                )

            return functional.drop_pixel_numpy(
                events=events, coordinates=self.coordinates
            )

        elif len(events.shape) == 4 or len(events.shape) == 3:
            if self.hot_pixel_frequency:
                self.coordinates = functional.identify_hot_pixel_raster(
                    events=events, hot_pixel_frequency=self.hot_pixel_frequency
                )

            return functional.drop_pixel_raster(events, self.coordinates)


@dataclass(frozen=True)
class Downsample:
    """Multiplies timestamps and spatial pixel coordinates with separate factors.
    Useful when the native temporal and/or spatial resolution of the original sensor is too
    high for downstream processing, notably when converting to dense representations of some sort.
    This transform does not drop any events.

    Parameters:
        time_factor (float): value to multiply timestamps with. Default is 1.
        spatial_factor (float): value to multiply pixel coordinates with. Default is 1.
                                Note that when using subsequential transforms that require
                                sensor_size, you must change the spatial values for the later
                                transformation.

    Example:
        >>> from tonic.transforms import Downsample
        >>> transform1 = Downsample(time_factor=0.001) # change us to ms
        >>> transform2 = Downsample(spatial_factor=0.25) # reduce focal plane to 1/4.
    """

    time_factor: float = 1
    spatial_factor: float = 1

    def __call__(self, events):
        events = events.copy()
        events = functional.time_skew_numpy(events, coefficient=self.time_factor)
        if "x" in events.dtype.names:
            events["x"] = events["x"] * self.spatial_factor
        if "y" in events.dtype.names:
            events["y"] = events["y"] * self.spatial_factor
        return events


@dataclass(frozen=True)
class MergePolarities:
    """
    Sets all polarities to zero. This transform does not have any parameters.

    Example:
        >>> transform = tonic.transforms.MergePolarities()
    """

    def __call__(self, events):
        events = events.copy()
        events["p"] = np.zeros_like(events["p"])
        return events


@dataclass(frozen=True)
class RandomCrop:
    """Crops the sensor size to a smaller size in a random location.

    x' = x - new_sensor_start_x

    y' = y - new_sensor_start_y

    Parameters:
        sensor_size: a 3-tuple of x,y,p for sensor_size
        target_size: a tuple of x,y target sensor size

    Example:
        >>> transform = tonic.transforms.RandomCrop(sensor_size=(340, 240, 2), target_size=(50, 50))
    """

    sensor_size: Tuple[int, int, int]
    target_size: Tuple[int, int]

    def __call__(self, events):

        return functional.crop_numpy(
            events=events, sensor_size=self.sensor_size, target_size=self.target_size
        )


@dataclass(frozen=True)
class RandomFlipPolarity:
    """Flips polarity of individual events with p.
    Changes polarities 1 to 0 and polarities [-1, 0] to 1

    Parameters:
        p (float): probability of flipping individual event polarities

    Example:
        >>> transform = tonic.transforms.RandomFlipPolarity(p=0.3)
    """

    p: float = 0.5

    def __post_init__(self):
        assert 0 <= self.p <= 1

    def __call__(self, events):
        events = events.copy()
        assert "p" in events.dtype.names
        if np.random.rand() <= self.p:
            events["p"] = np.invert(events["p"].astype(bool)).astype(events.dtype["p"])
        return events


@dataclass(frozen=True)
class RandomFlipLR:
    """Flips events in x with probability p. Pixels map as:

        x' = width - x

    Parameters:
        sensor_size: a 3-tuple of x,y,p for sensor_size
        p (float): probability of performing the flip

    Example:
        >>> transform = tonic.transforms.RandomFlipLR(p=0.3)
    """

    sensor_size: Tuple[int, int, int]
    p: float = 0.5

    def __post_init__(self):
        assert 0 <= self.p <= 1

    def __call__(self, events):
        events = events.copy()
        assert "x" in events.dtype.names
        if np.random.rand() <= self.p:
            events["x"] = self.sensor_size[0] - 1 - events["x"]
        return events


@dataclass(frozen=True)
class RandomFlipUD:
    """
    Flips events in y with probability p. Pixels map as:

        y' = height - y

    Parameters:
        sensor_size: a 3-tuple of x,y,p for sensor_size
        p (float): probability of performing the flip

    Example:
        >>> transform = tonic.transforms.RandomFlipUD(p=0.3)
    """

    sensor_size: Tuple[int, int, int]
    p: float = 0.5

    def __post_init__(self):
        assert 0 <= self.p <= 1

    def __call__(self, events):
        events = events.copy()
        assert "y" in events.dtype.names
        if np.random.rand() <= self.p:
            events["y"] = self.sensor_size[1] - 1 - events["y"]
        return events


@dataclass(frozen=True)
class RandomTimeReversal:
    """Reverses temporal order of events with probability p.

        .. math::
           t_i' = max(t) - t_i

    Parameters:
        p (float): probability of performing the flip
        flip_polarities (bool): if the time is reversed, also flip the polarities. True by default.

    Example:
        >>> transform = tonic.transforms.RandomTimeReversal(p=0.3)
    """

    p: float = 0.5
    flip_polarities: bool = True

    def __post_init__(self):
        assert 0 <= self.p <= 1

    def __call__(self, events):
        events = events.copy()
        assert "t" and "p" in events.dtype.names
        if np.random.rand() < self.p:
            events["t"] = np.max(events["t"]) - events["t"]
            if self.flip_polarities:
                events["p"] = np.invert(events["p"].astype(bool)).astype(
                    events.dtype["p"]
                )
            events = events[::-1]
        return events


@dataclass(frozen=True)
class RefractoryPeriod:
    """Sets a refractory period for each pixel, during which events will be
    ignored/discarded. We keep events if:

        .. math::
            t_n - t_{n-1} > t_{refrac}

    for each pixel.

    Parameters:
        delta (int): Refractory period for each pixel. Use same time
                     unit as event timestamps. Can use a 2-tuple to
                     sample from a range.

    >>> transform1 = tonic.transforms.RefractoryPeriod(delta=1000)
    >>> transform2 = tonic.transforms.RefractoryPeriod(delta=[0, 1000])
    """

    delta: Union[int, Tuple[int, int]]

    @staticmethod
    def get_params(delta: Union[int, Tuple[int, int]]):
        if type(delta) == tuple:
            delta = int((delta[1] - delta[0]) * np.random.random_sample() + delta[0])
        return delta

    def __call__(self, events):
        delta = self.get_params(delta=self.delta)
        return functional.refractory_period_numpy(
            events=events, refractory_period=delta
        )


@dataclass(frozen=True)
class SpatialJitter:
    """Changes x/y coordinate for each event by adding samples from a multivariate
    Gaussian distribution. It with the following properties:

        .. math::
            mean = [x,y]

            \Sigma = [[var_x, sigma_{xy}],[sigma_{xy}, var_y]]

    Jittered events that lie outside the focal plane will be dropped if clip_outliers is True.

    Parameters:
        sensor_size: a 3-tuple of x,y,p for sensor_size
        var_x (float): variance for the distribution in the x direction
        var_y (float): variance for the distribution in the y direction
        sigma_xy (float): changes skewness of distribution, only change if you want shifts along diagonal axis.
        clip_outliers (bool): when True, events that have been jittered outside the sensor size will be dropped.
    """

    sensor_size: Tuple[int, int, int]
    var_x: float = 1
    var_y: float = 1
    sigma_xy: float = 0
    clip_outliers: bool = False

    def __call__(self, events):
        events = events.copy()
        return functional.spatial_jitter_numpy(
            events=events,
            sensor_size=self.sensor_size,
            var_x=self.var_x,
            var_y=self.var_y,
            sigma_xy=self.sigma_xy,
            clip_outliers=self.clip_outliers,
        )


@dataclass
class TimeAlignment:
    """Removes offset for timestamps, so that first event starts at time zero."""

    def __call__(self, events):
        events = events.copy()
        assert "t" in events.dtype.names
        events["t"] -= min(events["t"])
        return events


@dataclass(frozen=True)
class TimeJitter:
    """Changes timestamp for each event by adding samples from a Gaussian
    distribution.

    Parameters:
        std (sequence or float): the standard deviation of the time jitter.
        clip_negative (bool): drops events that have negative timestamps.
        sort_timestamps (bool): sort the events by timestamps after jitter.
    """

    std: float
    clip_negative: bool = True
    sort_timestamps: bool = False

    def __call__(self, events):
        events = events.copy()
        return functional.time_jitter_numpy(
            events, self.std, self.clip_negative, self.sort_timestamps
        )


@dataclass(frozen=True)
class TimeSkew:
    """Skew all event timestamps according to a linear transform.

    Parameters:
        coefficient: a real-valued multiplier applied to the timestamps of the events.
                     E.g. a coefficient of 2.0 will double the effective delay between any
                     pair of events. Can provide a tuple for a range of values.
        offset: value by which the timestamps will be shifted after multiplication by
                the coefficient. Negative offsets are permissible but may result in
                in an exception if timestamps are shifted below 0. Tuple of values might
                be provided as a range to sample from.

    Example:
        >>> transform1 = TimeSkew(coefficient=1.3, offset=100)
        >>> transform2 = TimeSkew(coefficient=[0.8, 1.2], offset=[0, 150])
    """

    coefficient: Union[float, Tuple[float, float]]
    offset: Union[float, Tuple[float, float]] = 0

    def __call__(self, events):
        events = events.copy()
        return functional.time_skew_numpy(events, self.coefficient, self.offset)


@dataclass(frozen=True)
class UniformNoise:
    """
    Adds a fixed number of n noise events that are uniformly distributed
    across sensor size dimensions such as x, y, t and p.

    Parameters:
        sensor_size: a 3-tuple of x,y,p for sensor_size
        n: Number of events that are added. Can be a tuple of integers,
           so that n is sampled from a range.

    Example:
        >>> transform = tonic.transforms.UniformNoise(sensor_size=(340, 240, 2), n=3000)
    """

    sensor_size: Tuple[int, int, int]
    n: int

    @staticmethod
    def get_params(n: Union[int, Tuple[int, int]]):
        if type(n) == tuple:
            n = int((n[1] - n[0]) * np.random.random_sample() + n[0])
        return n

    def __call__(self, events):
        n = self.get_params(n=self.n)
        return functional.uniform_noise_numpy(
            events=events, sensor_size=self.sensor_size, n=n
        )


@dataclass(frozen=True)
class NumpyAsType:
    """
    Change dtype of numpy ndarray to custom dtype. This transform is necessary for example if you want
    to load raw events using a PyTorch dataloader. The original events coming from any dataset in Tonic
    are structured numpy arrays, so that they can be indexed as events["t"] or events["p"] etc. Pytorch's
    dataloader however does not support the conversion from structured numpy arrays to Tensors, that's
    why we need to employ at least NumpyAsType(int) to convert the structured array into an unstructured
    one before handing it to the dataloader.

    Parameters:
        dtype: data type that the array should be cast to.

    Example:
        >>> # indexing the dataset directly provides structured numpy arrays
        >>> dataset = tonic.datasets.NMNIST(save_to='data')
        >>> events, targets = dataset[100]
        >>>
        >>> # this doesn't work
        >>> dataloader = torch.utils.data.DataLoader(dataset)
        >>> events, targets = next(iter(dataloader))
        >>>
        >>> # we need to convert to unstructured arrays
        >>> transform = tonic.transforms.NumpyAsType(int)
        >>> dataset = tonic.datasets.NMNIST(save_to='data', transform=transform)
        >>> dataloader = torch.utils.data.DataLoader(dataset)
        >>> events, targets = next(iter(dataloader))
    """

    dtype: np.dtype

    def __call__(self, events):
        source_is_structured_array = (
            hasattr(events.dtype, "names") and events.dtype.names != None
        )
        target_is_structured_array = (
            hasattr(self.dtype, "names") and self.dtype.names != None
        )
        if source_is_structured_array and not target_is_structured_array:
            return np.lib.recfunctions.structured_to_unstructured(events, self.dtype)
        elif source_is_structured_array and target_is_structured_array:
            return NotImplementedError
        elif not target_is_structured_array and not source_is_structured_array:
            return events.astype(self.dtype)


@dataclass(frozen=True)
class ToAveragedTimesurface:
    """Create averaged timesurfaces for each event. Taken from the paper
    Sironi et al. 2018, HATS: Histograms of averaged time surfaces for robust event-based object classification
    https://openaccess.thecvf.com/content_cvpr_2018/papers/Sironi_HATS_Histograms_of_CVPR_2018_paper.pdf

    Parameters:
        sensor_size: a 3-tuple of x,y,p for sensor_size
        cell_size (int): size of each square in the grid
        surface_size (int): has to be odd
        temporal_window (float): how far back to look for past events for the time averaging
        tau (float): time constant to decay events around occuring event with.
        decay (str): can be either 'lin' or 'exp', corresponding to linear or exponential decay.
        num_workers (int): number of workers to be deployed on the histograms computation. When >1, joblib is required.
    """

    sensor_size: Tuple[int, int, int]
    cell_size: int = 10
    surface_size: int = 7
    temporal_window: int = 5e5
    tau: int = 5e3
    decay: str = "lin"
    num_workers: int = 1

    def __call__(self, events):
        return functional.to_averaged_timesurface(
            events,
            sensor_size=self.sensor_size,
            cell_size=self.cell_size,
            surface_size=self.surface_size,
            temporal_window=self.temporal_window,
            tau=self.tau,
            decay=self.decay,
            num_workers=self.num_workers,
        )


@dataclass(frozen=True)
class ToFrame:
    """Accumulate events to frames by slicing along constant time (time_window),
    constant number of events (spike_count) or constant number of frames (n_time_bins / n_event_bins).
    All the events in one slice are added up in a frame for each polarity.
    You can set one of the first 4 parameters to choose the slicing method. Depending on which method you choose,
    overlap will assume different functionality, whether that might be temporal overlap, number of events
    or fraction of a bin. As a rule of thumb, here are some considerations if you are unsure which slicing
    method to choose:

    * If your recordings are of roughly the same length, a safe option is to set time_window. Bare in mind
      that the number of events can vary greatly from slice to slice, but will give you some consistency when
      training RNNs or other algorithms that have time steps.

    * If your recordings have roughly the same amount of activity / number of events and you are more interested
      in the spatial composition, then setting spike_count will give you frames that are visually more consistent.

    * The previous time_window and spike_count methods will likely result in a different amount of frames for each
      recording. If your training method benefits from consistent number of frames across a dataset (for easier
      batching for example), or you want a parameter that is easier to set than the exact window length or number
      of events per slice, consider fixing the number of frames by setting n_time_bins or n_event_bins. The two
      methods slightly differ with respect to how the slices are distributed across the recording. You can define
      an overlap between 0 and 1 to provide some robustness.

    Parameters:
        sensor_size: a 3-tuple of x,y,p for sensor_size. If omitted, the sensor size is calculated for that sample. However,
                    do use this feature sparingly as when not all pixels fire in a sample, this might cause issues with batching/
                    stacking tensors further down the line.
        time_window (float): time window length for one frame. Use the same time unit as timestamps in the event recordings.
                             Good if you want temporal consistency in your training, bad if you need some visual consistency
                             for every frame if the recording's activity is not consistent.
        spike_count (int): number of events per frame. Good for training CNNs which do not care about temporal consistency.
        n_time_bins (int): fixed number of frames, sliced along time axis. Good for generating a pre-determined number of
                           frames which might help with batching.
        n_event_bins (int): fixed number of frames, sliced along number of events in the recording. Good for generating a
                            pre-determined number of frames which might help with batching.
        overlap (float): overlap between frames defined either in time units, number of events or number of bins between 0 and 1.
        include_incomplete (bool): if True, includes overhang slice when time_window or spike_count is specified.
                                   Not valid for bin_count methods.

    Example:
        >>> from tonic.transforms import ToFrame
        >>> transform1 = ToFrame(time_window=10000, overlap=300, include_incomplete=True)
        >>> transform2 = ToFrame(spike_count=3000, overlap=100, include_incomplete=True)
        >>> transform3 = ToFrame(n_time_bins=100, overlap=0.1)
    """

    sensor_size: Optional[Tuple[int, int, int]]
    time_window: Optional[float] = None
    event_count: Optional[int] = None
    n_time_bins: Optional[int] = None
    n_event_bins: Optional[int] = None
    overlap: float = 0
    include_incomplete: bool = False

    def __call__(self, events):

        return functional.to_frame_numpy(
            events=events,
            sensor_size=self.sensor_size,
            time_window=self.time_window,
            event_count=self.event_count,
            n_time_bins=self.n_time_bins,
            n_event_bins=self.n_event_bins,
            overlap=self.overlap,
            include_incomplete=self.include_incomplete,
        )


@dataclass(frozen=True)
class ToSparseTensor:
    """
    PyTorch sparse tensor drop-in replacement for ToFrame. See https://pytorch.org/docs/stable/sparse.html for details
    about sparse tensors. The dense shape of the tensor will be (TCWH) and can be inflated by calling to_dense().
    You need to have PyTorch installed for this transformation. Under the hood this transform calls ToFrame() with the
    same parameters, converts to a pytorch tensor and calls to_sparse().

    Parameters:
        sensor_size: a 3-tuple of x,y,p for sensor_size. If omitted, the sensor size is calculated for that sample. However,
                    do use this feature sparingly as when not all pixels fire in a sample, this might cause issues with batching/
                    stacking tensors further down the line.
        time_window (float): time window length for one frame. Use the same time unit as timestamps in the event recordings.
                             Good if you want temporal consistency in your training, bad if you need some visual consistency
                             for every frame if the recording's activity is not consistent.
        spike_count (int): number of events per frame. Good for training CNNs which do not care about temporal consistency.
        n_time_bins (int): fixed number of frames, sliced along time axis. Good for generating a pre-determined number of
                           frames which might help with batching.
        n_event_bins (int): fixed number of frames, sliced along number of events in the recording. Good for generating a
                            pre-determined number of frames which might help with batching.
        overlap (float): overlap between frames defined either in time units, number of events or number of bins between 0 and 1.
        include_incomplete (bool): if True, includes overhang slice when time_window or spike_count is specified.
                                   Not valid for bin_count methods.

    Example:
        >>> from tonic.transforms import ToSparseTensor
        >>> transform1 = ToSparseTensor(time_window=10000, overlap=300, include_incomplete=True)
        >>> transform2 = ToSparseTensor(spike_count=3000, overlap=100, include_incomplete=True)
        >>> transform3 = ToSparseTensor(n_time_bins=100, overlap=0.1)
    """

    sensor_size: Tuple[int, int, int]
    time_window: Optional[float] = None
    event_count: Optional[int] = None
    n_time_bins: Optional[int] = None
    n_event_bins: Optional[int] = None
    overlap: float = 0
    include_incomplete: bool = False

    def __call__(self, events):
        try:
            import torch
        except ImportError:
            raise ImportError("PyTorch not installed.")

        dense_frames = functional.to_frame_numpy(
            events=events,
            sensor_size=self.sensor_size,
            time_window=self.time_window,
            event_count=self.event_count,
            n_time_bins=self.n_time_bins,
            n_event_bins=self.n_event_bins,
            overlap=self.overlap,
            include_incomplete=self.include_incomplete,
        )
        return torch.from_numpy(dense_frames).to_sparse()


@dataclass(frozen=True)
class ToImage:
    """Counts up all events to a *single* image of size sensor_size. ToImage will typically
    be used in combination with SlicedDataset to cut a recording into smaller chunks that
    are then individually binned to frames.
    """

    sensor_size: Tuple[int, int, int]

    def __call__(self, events):

        frames = functional.to_frame_numpy(
            events=events, sensor_size=self.sensor_size, event_count=len(events)
        )

        return frames.squeeze(0)


@dataclass(frozen=True)
class ToTimesurface:
    """Create global or local time surfaces for each event. Modeled after the paper
    Lagorce et al. 2016, Hots: a hierarchy of event-based time-surfaces for pattern recognition
    https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=7508476

    Parameters:
        sensor_size: a 3-tuple of x,y,p for sensor_size
        surface_dimensions (int, int): width does not have to be equal to height, however both numbers have to be odd.
            if surface_dimensions is None: the time surface is defined globally, on the whole sensor grid.
        tau (float): time constant to decay events around occuring event with.
        decay (str): can be either 'lin' or 'exp', corresponding to linear or exponential decay.
    """

    sensor_size: Tuple[int, int, int]
    surface_dimensions: Union[None, Tuple[int, int]] = None
    tau: float = 5e3
    decay: str = "lin"

    def __call__(self, events):

        return functional.to_timesurface_numpy(
            events=events,
            sensor_size=self.sensor_size,
            surface_dimensions=self.surface_dimensions,
            tau=self.tau,
            decay=self.decay,
        )


@dataclass(frozen=True)
class ToVoxelGrid:
    """Build a voxel grid with bilinear interpolation in the time domain from a set of events.
    Implements the event volume from Zhu et al. 2019, Unsupervised event-based learning
    of optical flow, depth, and egomotion.

    Parameters:
        sensor_size: a 3-tuple of x,y,p for sensor_size
        n_time_bins (int): fixed number of time bins to slice the event sample into."""

    sensor_size: Tuple[int, int, int]
    n_time_bins: int

    def __call__(self, events):

        return functional.to_voxel_grid_numpy(
            events.copy(), self.sensor_size, self.n_time_bins
        )


@dataclass(frozen=True)
class ToBinaRep:
    """Takes T*B binary event frames to produce a sequence of T frames of N-bit numbers.
    To do so, N binary frames are interpreted as a single frame of N-bit representation. Taken from the paper
    Barchid et al. 2022, Bina-Rep Event Frames: a Simple and Effective Representation for Event-based cameras
    https://arxiv.org/pdf/2202.13662.pdf

    Parameters:
        n_frames (int): the number T of bina-rep frames.
        n_bits (int): the number N of bits used in the N-bit representation.


    Example:
        >>> n_time_bins = n_frames * n_bits
        >>>
        >>> transforms.Compose([
        >>>     transforms.ToFrame(
        >>>         sensor_size=sensor_size,
        >>>         n_time_bins=n_time_bins,
        >>>     ),
        >>>     transforms.ToBinaRep(
        >>>         n_frames=n_frames,
        >>>         n_bits=n_bits,
        >>>     ),
        >>> ])
    """

    n_frames: Optional[int] = 1
    n_bits: Optional[int] = 8

    def __call__(self, event_frames):

        return functional.to_bina_rep_numpy(event_frames, self.n_frames, self.n_bits)


@dataclass(frozen=True)
class Repeat:
    """Copies target n times. Useful to transform sample labels into sequences."""

    n_repeat: int

    def __call__(self, target):
        return np.tile(np.expand_dims(target, 0), [self.n_repeat, 1])


@dataclass(frozen=True)
class ToOneHotEncoding:
    """Transforms one or more targets into a one hot encoding scheme."""

    n_classes: int

    def __call__(self, target):
        return np.eye(self.n_classes)[target]



"""
==================
TimeJitter
==================
The :class:`~tonic.transforms.TimeJitter` jitters timestamps.
"""

import tonic

nmnist = tonic.datasets.NMNIST("../../tutorials/data", train=False)
events, label = nmnist[0]

transform = tonic.transforms.Compose(
    [
        tonic.transforms.TimeJitter(std=100, clip_negative=True),
        tonic.transforms.ToFrame(
            sensor_size=nmnist.sensor_size,
            time_window=10000,
        ),
    ]
)

frames = transform(events)

ani = tonic.utils.plot_animation(frames)



"""
=========
ToBinaRep
=========

The :class:`~tonic.transforms.ToBinaRep` creates binary representations of frames.
"""

import tonic
import tonic.transforms as transforms

nmnist = tonic.datasets.NMNIST("../../tutorials/data", train=False)
events, label = nmnist[0]

transform = transforms.Compose(
    [
        transforms.ToFrame(
            sensor_size=nmnist.sensor_size,
            n_time_bins=30 * 2,
        ),
        transforms.ToBinaRep(n_frames=30, n_bits=2),
    ]
)

frames = transform(events)

ani = tonic.utils.plot_animation(frames)



import tonic
import pytest


def test_train_split():
    dataset = tonic.datasets.DSEC(
        save_to="data",
        split="train",
        data_selection=[],
    )
    assert len(dataset) == 41


def test_test_split():
    dataset = tonic.datasets.DSEC(
        save_to="data",
        split="test",
        data_selection=[],
    )
    assert len(dataset) == 12


def test_optical_flow_subset():
    dataset = tonic.datasets.DSEC(
        save_to="data",
        split="train",
        data_selection=[],
        target_selection="optical_flow_forward_timestamps",
    )
    assert len(dataset) == 18
    data, targets = dataset[0]
    assert len(data) == 0
    assert len(targets) == 1


def test_data_selection():
    dataset = tonic.datasets.DSEC(
        save_to="data", split="thun_00_a", data_selection="image_timestamps"
    )
    data, targets = dataset[0]
    assert len(data) == 1
    assert len(targets) == 0


def test_multi_data_selection():
    dataset = tonic.datasets.DSEC(
        save_to="data",
        split="thun_00_a",
        data_selection=["image_timestamps", "image_exposure_timestamps_left"],
    )
    data, targets = dataset[0]
    assert len(data) == 2
    assert len(targets) == 0


def test_target_selection():
    dataset = tonic.datasets.DSEC(
        save_to="data",
        split="thun_00_a",
        data_selection=[],
        target_selection="disparity_timestamps",
    )
    data, targets = dataset[0]
    assert len(data) == 0
    assert len(targets) == 1


def test_target_multiselection():
    dataset = tonic.datasets.DSEC(
        save_to="data",
        split="thun_00_a",
        data_selection="image_timestamps",
        target_selection=["disparity_timestamps", "optical_flow_forward_timestamps"],
    )
    data, targets = dataset[0]
    assert len(data) == 1
    assert len(targets) == 2


def test_optical_flow():
    dataset = tonic.datasets.DSEC(
        save_to="data",
        split="thun_00_a",
        data_selection="image_timestamps",
        target_selection=[
            "optical_flow_forward_event",
            "optical_flow_forward_timestamps",
        ],
    )
    data, targets = dataset[0]
    assert len(data) == 1
    assert len(targets) == 2


def test_raises_exception_wrong_recording_name():
    with pytest.raises(Exception):
        tonic.datasets.DSEC(
            save_to="data", split="wrong_name", data_selection=["image_timestamps"]
        )


def test_raises_exception_data_name():
    with pytest.raises(Exception):
        tonic.datasets.DSEC(
            save_to="data", split="thun_00_a", data_selection="wrong_data"
        )


def test_raises_exception_wrong_recording_name():
    with pytest.raises(Exception):
        tonic.datasets.DSEC(
            save_to="data",
            split="thun_00_a",
            data_selection=[],
            target_selection="wrong_target",
        )


def test_raises_exception_combination_test_train():
    with pytest.raises(Exception):
        tonic.datasets.DSEC(
            save_to="data",
            split=["zurich_city_11_c", "thun_01_a"],
            data_selection=["image_timestamps"],
        )


def test_raises_exception_targets_for_test():
    with pytest.raises(Exception):
        tonic.datasets.DSEC(
            save_to="data",
            split="test",
            data_selection=["image_timestamps"],
            target_selection="disparity_timestamps",
        )



import tonic
import tonic.transforms as transforms
from utils import create_random_input
from sys import platform
import pytest


class DummyDataset:
    def __init__(self, events, transform):
        self.events = events
        self.transform = transform

    def __getitem__(self, index):
        return self.transform(self.events[index]), 1

    def __len__(self):
        return len(self.events)


def test_pytorch_batch_collation_dense_tensor():
    import torch

    events1, sensor_size = create_random_input()
    events2, sensor_size = create_random_input()

    time_window = 1000
    transform = transforms.Compose(
        [transforms.ToFrame(sensor_size=sensor_size, time_window=time_window)]
    )
    dataset = DummyDataset(
        (events1[:5000], events2), transform
    )  # simulate recordings of different length
    batch_size = 2
    dataloader = torch.utils.data.DataLoader(
        dataset,
        collate_fn=tonic.collation.PadTensors(batch_first=False),
        batch_size=batch_size,
    )

    batch, label = next(iter(dataloader))

    max_time = int(events2["t"][-1])
    assert batch.shape[0] == max_time // time_window
    assert batch.shape[1] == batch_size
    assert batch.shape[2] == sensor_size[2]


@pytest.mark.skipif(not platform == "linux", reason="Test doesn't finish on Windows.")
def test_plotting():
    events, sensor_size = create_random_input()
    tonic.utils.plot_event_grid(events)


@pytest.mark.skipif(not platform == "linux", reason="Test doesn't finish on Windows.")
def test_animation():
    events, sensor_size = create_random_input()

    transform = tonic.transforms.ToFrame(
        sensor_size=sensor_size,
        time_window=100000,
    )

    frames = transform(events)
    tonic.utils.plot_animation(frames)



import numpy as np
import tonic.datasets as datasets
import dataset_utils
import os
from utils import create_random_input
from tonic.download_utils import download_url


base_url = "https://www.neuromorphic-vision.com/public/downloads/dataset_samples/"


class ASLDVSTestCaseTest(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.ASLDVS
    FEATURE_TYPES = (datasets.ASLDVS.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "ASLDVS/a")
        os.makedirs(testfolder, exist_ok=True)
        filename = "a_0244.mat"
        download_url(url=base_url + filename, root=testfolder, filename=filename)
        return {"n_samples": 1}


class DVSGestureTestCaseTrain(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.DVSGesture
    FEATURE_TYPES = (datasets.DVSGesture.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "DVSGesture/ibmGestureTrain/user24_led")
        os.makedirs(testfolder, exist_ok=True)
        events, sensor_size = create_random_input(dtype=datasets.DVSGesture.dtype)
        events = np.lib.recfunctions.structured_to_unstructured(events)
        np.save(testfolder + "/0.npy", events)
        np.save(testfolder + "/1.npy", events)
        return {"n_samples": 2}


class DVSGestureTestCaseTest(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.DVSGesture
    FEATURE_TYPES = (datasets.DVSGesture.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "DVSGesture/ibmGestureTest/user24_led")
        os.makedirs(testfolder, exist_ok=True)
        events, sensor_size = create_random_input(dtype=datasets.DVSGesture.dtype)
        events = np.lib.recfunctions.structured_to_unstructured(events)
        np.save(testfolder + "/0.npy", events)
        return {"n_samples": 1}


class NCaltech101TestCase(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NCALTECH101
    FEATURE_TYPES = (datasets.NCALTECH101.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "NCALTECH101/Caltech101/airplanes/")
        os.makedirs(testfolder, exist_ok=True)
        filename = "image_0006.bin"
        download_url(url=base_url + filename, root=testfolder, filename=filename)
        return {"n_samples": 1}


class NMNISTTestCaseTrain(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NMNIST
    FEATURE_TYPES = (datasets.NMNIST.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "NMNIST/Train/1/")
        os.makedirs(testfolder, exist_ok=True)
        filename = "image_0006.bin"
        download_url(url=base_url + filename, root=testfolder, filename="24901.bin")
        return {"n_samples": 1}


class NMNISTTestCaseTest(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NMNIST
    FEATURE_TYPES = (datasets.NMNIST.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "NMNIST/Test/1/")
        os.makedirs(testfolder, exist_ok=True)
        filename = "image_0006.bin"
        download_url(url=base_url + filename, root=testfolder, filename="04652.bin")
        return {"n_samples": 1}


class DVSLipTestCaseTrain(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.DVSLip
    FEATURE_TYPES = (datasets.DVSLip.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "DVSLip/DVS-Lip/train/accused")
        os.makedirs(testfolder, exist_ok=True)
        events, sensor_size = create_random_input(
            dtype=np.dtype([("t", "<i4"), ("x", "i1"), ("y", "i1"), ("p", "i1")])
        )
        np.save(testfolder + "/0.npy", events)
        np.save(testfolder + "/1.npy", events)
        return {"n_samples": 2}


class DVSLipTestCaseTest(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.DVSLip
    FEATURE_TYPES = (datasets.DVSLip.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "DVSLip/DVS-Lip/test/accused")
        os.makedirs(testfolder, exist_ok=True)
        events, sensor_size = create_random_input(
            dtype=np.dtype([("t", "<i4"), ("x", "i1"), ("y", "i1"), ("p", "i1")])
        )
        np.save(testfolder + "/0.npy", events)
        return {"n_samples": 1}
